import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.named_tuples import CfxKeyData, Rfc1964KeyData
    from gssapi.raw.sec_contexts import SecurityContext


class Krb5LucidContext:
    """
    The base container returned by :meth:`krb5_export_lucid_sec_context` when
    an unknown version was requested.
    """


class Krb5LucidContextV1(Krb5LucidContext):
    """
    Kerberos context data returned by :meth:`krb5_export_lucid_sec_context`
    when version 1 was requested.
    """

    @property
    def version(self) -> t.Optional[int]:
        """The structure version number

        Returns:
            Optional[int]: the structure version number
        """

    @property
    def is_initiator(self) -> t.Optional[bool]:
        """Whether the context was the initiator

        Returns:
            Optional[bool]: ``True`` when the exported context was the
            initiator
        """

    @property
    def endtime(self) -> t.Optional[int]:
        """Expiration time of the context

        Returns:
            Optional[int]: the expiration time of the context
        """

    @property
    def send_seq(self) -> t.Optional[int]:
        """Sender sequence number

        Returns:
            Optional[int]: the sender sequence number
        """

    @property
    def recv_seq(self) -> t.Optional[int]:
        """Receiver sequence number

        Returns:
            Optional[int]: the receiver sequence number
        """

    @property
    def protocol(self) -> t.Optional[int]:
        """The protocol number

        If the protocol number is 0 then :attr:`rfc1964_kd` is set and
        :attr:`cfx_kd` is `None`. If the protocol number is 1 then the opposite
        is true.

        Protocol 0 refers to RFC1964 and 1 refers to RFC4121.

        Returns:
            Optional[int]: the protocol number
        """

    @property
    def rfc1964_kd(self) -> t.Optional["Rfc1964KeyData"]:
        """Keydata for protocol 0 (RFC1964)

        This will be set when :attr:`protocol` is ``0``.

        Returns:
            Optional[Rfc1964KeyData]: the RFC1964 key data
        """

    @property
    def cfx_kd(self) -> t.Optional["CfxKeyData"]:
        """Key data for protocol 1 (RFC4121)

        This will be set when :attr:`protocol` is ``1``.

        Returns:
            Optional[CfxKeyData]: the RFC4121 key data
        """


def krb5_ccache_name(
    name: t.Optional[bytes],
) -> bytes:
    """Set the default Kerberos Protocol credentials cache name.

    This method sets the default credentials cache name for use by he Kerberos
    mechanism. The default credentials cache is used by
    :meth:`~gssapi.raw.creds.acquire_cred` to create a GSS-API credential. It
    is also used by :meth:`~gssapi.raw.sec_contexts.init_sec_context` when
    `GSS_C_NO_CREDENTIAL` is specified.

    Note:
        Heimdal does not return the old name when called. It also does not
        reset the ccache lookup behaviour when setting to ``None``.

    Note:
        The return value may not be thread safe.

    Args:
        name (Optional[bytes]): the name to set as the new thread specific
            ccache name. Set to ``None`` to revert back to getting the ccache
            from the config/environment settings.

    Returns:
        bytes: the old name that was previously set

    Raises:
        ~gssapi.exceptions.GSSError
    """


def krb5_export_lucid_sec_context(
    context: "SecurityContext",
    version: int,
) -> Krb5LucidContext:
    """Returns a non-opaque version of the internal context info.

    Gets information about the Kerberos security context passed in. Currently
    only version 1 is known and supported by this library.

    Note:
        The context handle must not be used again by the caller after this
        call.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        version (int): the output structure version to export.  Currently
            only 1 is supported.

    Returns:
        Krb5LucidContext: the non-opaque version context info

    Raises:
        ~gssapi.exceptions.GSSError
    """


def krb5_extract_authtime_from_sec_context(
    context: "SecurityContext",
) -> int:
    """Get the auth time for the security context.

    Gets the auth time for the established security context.

    Note:
        Heimdal can only get the authtime on the acceptor security context.
        MIT is able to get the authtime on both initiators and acceptors.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context

    Returns:
        int: the authtime

    Raises:
        ~gssapi.exceptions.GSSError
    """


def krb5_extract_authz_data_from_sec_context(
    context: "SecurityContext",
    ad_type: int,
) -> bytes:
    """Extracts Kerberos authorization data.

    Extracts authorization data that may be stored within the context.

    Note:
        Only operates on acceptor contexts.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        ad_type (int): the type of data to extract

    Returns:
        bytes: the raw authz data from the sec context

    Raises:
        ~gssapi.exceptions.GSSError
    """


def krb5_import_cred(
    cred_handle: "Creds",
    cache: t.Optional[int] = None,
    keytab_principal: t.Optional[int] = None,
    keytab: t.Optional[int] = None,
) -> None:
    """Import Krb5 credentials into GSSAPI credential.

    Imports the krb5 credentials (either or both of the keytab and cache) into
    the GSSAPI credential so it can be used within GSSAPI. The ccache is
    copied by reference and thus shared, so if the credential is destroyed,
    all users of cred_handle will fail.

    Args:
        cred_handle (Creds): the credential handle to import into
        cache (int): the krb5_ccache address pointer, as an int, to import
            from
        keytab_principal (int): the krb5_principal address pointer, as an int,
            of the credential to import
        keytab (int): the krb5_keytab address pointer, as an int, of the
            keytab to import

    Returns:
        None

    Raises:
        ~gssapi.exceptions.GSSError
    """


def krb5_get_tkt_flags(
    context: "SecurityContext",
) -> int:
    """Return ticket flags for the kerberos ticket.

    Return the ticket flags for the kerberos ticket received when
    authenticating the initiator.

    Note:
        Heimdal can only get the tkt flags on the acceptor security context.
        MIT is able to get the tkt flags on initiators and acceptors.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the security
            context

    Returns:
        int: the ticket flags for the received kerberos ticket

    Raises:
        ~gssapi.exceptions.GSSError
    """


def krb5_set_allowable_enctypes(
    cred_handle: "Creds",
    ktypes: t.Iterable[int],
) -> None:
    """Limits the keys that can be exported.

    Called by a context initiator after acquiring the creds but before calling
    :meth:`~gssapi.raw.sec_contexts.init_sec_context` to restrict the set of
    enctypes which will be negotiated during context establisment to those in
    the provided list.

    Warning:
        The cred_handle should not be ``GSS_C_NO_CREDENTIAL``.

    Args:
        cred_hande (Creds): the credential handle
        ktypes (List[int]): list of enctypes allowed

    Returns:
        None

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

from gssapi.raw import creds as rcreds
from gssapi.raw import named_tuples as tuples
from gssapi.raw import names as rnames
from gssapi.raw import oids as roids
from gssapi._utils import import_gssapi_extension, _encode_dict

from gssapi import names

rcred_imp_exp = import_gssapi_extension('cred_imp_exp')
rcred_s4u = import_gssapi_extension('s4u')
rcred_cred_store = import_gssapi_extension('cred_store')
rcred_rfc5588 = import_gssapi_extension('rfc5588')


class Credentials(rcreds.Creds):
    """GSSAPI Credentials

    This class represents a set of GSSAPI credentials which may
    be used with and/or returned by other GSSAPI methods.

    It inherits from the low-level GSSAPI :class:`~gssapi.raw.creds.Creds`
    class, and thus may used with both low-level and high-level API methods.

    If your implementation of GSSAPI supports the credentials import-export
    extension, you may pickle and unpickle this object.

    The constructor either acquires or imports a set of GSSAPI
    credentials.

    If the `base` argument is used, an existing
    :class:`~gssapi.raw.creds.Creds` object from the low-level API is
    converted into a high-level object.

    If the `token` argument is used, the credentials
    are imported using the token, if the credentials import-export
    extension is supported (:requires-ext:`cred_imp_exp`).

    Otherwise, the credentials are acquired as per the
    :meth:`acquire` method.

    Raises:
        ~gssapi.exceptions.BadMechanismError
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
        ~gssapi.exceptions.ExpiredCredentialsError
        ~gssapi.exceptions.MissingCredentialsError
    """

    __slots__ = ()

    def __new__(
        cls,
        base: t.Optional[rcreds.Creds] = None,
        token: t.Optional[bytes] = None,
        name: t.Optional[rnames.Name] = None,
        lifetime: t.Optional[int] = None,
        mechs: t.Optional[t.Iterable[roids.OID]] = None,
        usage: str = 'both',
        store: t.Optional[
            t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
        ] = None,
    ) -> "Credentials":
        # TODO(directxman12): this is missing support for password
        #                     (non-RFC method)
        if base is not None:
            base_creds = base
        elif token is not None:
            if rcred_imp_exp is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not have support for importing and "
                                          "exporting creditials")

            base_creds = rcred_imp_exp.import_cred(token)
        else:
            res = cls.acquire(name, lifetime, mechs, usage,
                              store=store)
            base_creds = res.creds

        return t.cast("Credentials",
                      super(Credentials, cls).__new__(cls, base_creds))

    @property
    def name(self) -> names.Name:
        """Get the name associated with these credentials"""
        return t.cast(names.Name,
                      self.inquire(name=True, lifetime=False, usage=False,
                                   mechs=False).name)

    @property
    def lifetime(self) -> int:
        """Get the remaining lifetime of these credentials, in seconds"""
        return t.cast(int,
                      self.inquire(name=False, lifetime=True,
                                   usage=False, mechs=False).lifetime)

    @property
    def mechs(self) -> t.Set[roids.OID]:
        """Get the mechanisms for these credentials"""
        return t.cast(t.Set[roids.OID],
                      self.inquire(name=False, lifetime=False,
                                   usage=False, mechs=True).mechs)

    @property
    def usage(self) -> str:
        """Get the usage (initiate, accept, or both) of these credentials"""
        return t.cast(str,
                      self.inquire(name=False, lifetime=False,
                                   usage=True, mechs=False).usage)

    @classmethod
    def acquire(
        cls,
        name: t.Optional[rnames.Name] = None,
        lifetime: t.Optional[int] = None,
        mechs: t.Optional[t.Iterable[roids.OID]] = None,
        usage: str = 'both',
        store: t.Optional[
            t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
        ] = None,
    ) -> tuples.AcquireCredResult:
        """Acquire GSSAPI credentials

        This method acquires credentials.  If the `store` argument is
        used, the credentials will be acquired from the given
        credential store (if supported).  Otherwise, the credentials are
        acquired from the default store.

        The credential store information is a dictionary containing
        mechanisms-specific keys and values pointing to a credential store
        or stores.

        Using a non-default store requires support for the credentials store
        extension.

        Args:
            name (~gssapi.names.Name): the name associated with the
                credentials, or None for the default name
            lifetime (int): the desired lifetime of the credentials in seconds,
                or None for indefinite
            mechs (list): the desired :class:`MechType` OIDs to be used
                with the credentials, or None for the default set
            usage (str): the usage for the credentials -- either 'both',
                'initiate', or 'accept'
            store (dict): the credential store information pointing to the
                credential store from which to acquire the credentials,
                or None for the default store (:requires-ext:`cred_store`)

        Returns:
            AcquireCredResult: the acquired credentials and information about
            them

        Raises:
            ~gssapi.exceptions.BadMechanismError
            ~gssapi.exceptions.BadNameTypeError
            ~gssapi.exceptions.BadNameError
            ~gssapi.exceptions.ExpiredCredentialsError
            ~gssapi.exceptions.MissingCredentialsError
        """

        if store is None:
            res = rcreds.acquire_cred(name, lifetime,
                                      mechs, usage)
        else:
            if rcred_cred_store is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not have support for manipulating "
                                          "credential stores")

            b_store = _encode_dict(store)

            res = rcred_cred_store.acquire_cred_from(b_store, name,
                                                     lifetime, mechs,
                                                     usage)

        return tuples.AcquireCredResult(cls(base=res.creds), res.mechs,
                                        res.lifetime)

    def store(
        self,
        store: t.Optional[
            t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
        ] = None,
        usage: str = 'both',
        mech: t.Optional[roids.OID] = None,
        overwrite: bool = False,
        set_default: bool = False,
    ) -> tuples.StoreCredResult:
        """Store these credentials into the given store

        This method stores the current credentials into the specified
        credentials store.  If the default store is used, support for
        :rfc:`5588` is required.  Otherwise, support for the credentials
        store extension is required.

        :requires-ext:`rfc5588` or :requires-ext:`cred_store`

        Args:
            store (dict): the store into which to store the credentials,
                or None for the default store.
            usage (str): the usage to store the credentials with -- either
                'both', 'initiate', or 'accept'
            mech (~gssapi.OID): the :class:`MechType` to associate with the
                stored credentials
            overwrite (bool): whether or not to overwrite existing credentials
                stored with the same name, etc
            set_default (bool): whether or not to set these credentials as
                the default credentials for the given store.

        Returns:
            StoreCredResult: the results of the credential storing operation

        Raises:
            ~gssapi.exceptions.GSSError
            ~gssapi.exceptions.ExpiredCredentialsError
            ~gssapi.exceptions.MissingCredentialsError
            ~gssapi.exceptions.OperationUnavailableError
            ~gssapi.exceptions.DuplicateCredentialsElementError
        """

        if store is None:
            if rcred_rfc5588 is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not have support for RFC 5588")

            return rcred_rfc5588.store_cred(self, usage, mech,
                                            overwrite, set_default)
        else:
            if rcred_cred_store is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not have support for manipulating "
                                          "credential stores directly")

            b_store = _encode_dict(store)

            return rcred_cred_store.store_cred_into(b_store, self, usage, mech,
                                                    overwrite, set_default)

    def impersonate(
        self,
        name: t.Optional[rnames.Name] = None,
        lifetime: t.Optional[int] = None,
        mechs: t.Optional[t.Iterable[roids.OID]] = None,
        usage: str = 'initiate',
    ) -> "Credentials":
        """Impersonate a name using the current credentials

        This method acquires credentials by impersonating another
        name using the current credentials.

        :requires-ext:`s4u`

        Args:
            name (~gssapi.names.Name): the name to impersonate
            lifetime (int): the desired lifetime of the new credentials in
                seconds, or None for indefinite
            mechs (list): the desired :class:`MechType` OIDs for the new
                credentials
            usage (str): the desired usage for the new credentials -- either
                'both', 'initiate', or 'accept'.  Note that some mechanisms
                may only support 'initiate'.

        Returns:
            Credentials: the new credentials impersonating the given name
        """

        if rcred_s4u is None:
            raise NotImplementedError("Your GSSAPI implementation does not "
                                      "have support for S4U")

        res = rcred_s4u.acquire_cred_impersonate_name(self, name,
                                                      lifetime, mechs,
                                                      usage)

        return type(self)(base=res.creds)

    def inquire(
        self,
        name: bool = True,
        lifetime: bool = True,
        usage: bool = True,
        mechs: bool = True,
    ) -> tuples.InquireCredResult:
        """Inspect these credentials for information

        This method inspects these credentials for information about them.

        Args:
            name (bool): get the name associated with the credentials
            lifetime (bool): get the remaining lifetime for the credentials
            usage (bool): get the usage for the credentials
            mechs (bool): get the mechanisms associated with the credentials

        Returns:
            InquireCredResult: the information about the credentials,
            with None used when the corresponding argument was False

        Raises:
            ~gssapi.exceptions.MissingCredentialsError
            ~gssapi.exceptions.InvalidCredentialsError
            ~gssapi.exceptions.ExpiredCredentialsError
        """

        res = rcreds.inquire_cred(self, name, lifetime, usage, mechs)

        if res.name is not None:
            res_name = names.Name(res.name)
        else:
            res_name = None

        return tuples.InquireCredResult(res_name, res.lifetime,
                                        res.usage, res.mechs)

    def inquire_by_mech(
        self,
        mech: roids.OID,
        name: bool = True,
        init_lifetime: bool = True,
        accept_lifetime: bool = True,
        usage: bool = True,
    ) -> tuples.InquireCredByMechResult:
        """Inspect these credentials for per-mechanism information

        This method inspects these credentials for per-mechanism information
        about them.

        Args:
            mech (~gssapi.OID): the mechanism for which to retrieve the
                information
            name (bool): get the name associated with the credentials
            init_lifetime (bool): get the remaining initiate lifetime for
                the credentials in seconds
            accept_lifetime (bool): get the remaining accept lifetime for
                the credentials in seconds
            usage (bool): get the usage for the credentials

        Returns:
            InquireCredByMechResult: the information about the credentials,
            with None used when the corresponding argument was False
        """

        res = rcreds.inquire_cred_by_mech(self, mech, name, init_lifetime,
                                          accept_lifetime, usage)

        if res.name is not None:
            res_name = names.Name(res.name)
        else:
            res_name = None

        return tuples.InquireCredByMechResult(res_name,
                                              res.init_lifetime,
                                              res.accept_lifetime,
                                              res.usage)

    def add(
        self,
        name: rnames.Name,
        mech: roids.OID,
        usage: str = 'both',
        init_lifetime: t.Optional[int] = None,
        accept_lifetime: t.Optional[int] = None,
        impersonator: t.Optional[rcreds.Creds] = None,
        store: t.Optional[
            t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
        ] = None,
    ) -> "Credentials":
        """Acquire more credentials to add to the current set

        This method works like :meth:`acquire`, except that it adds the
        acquired credentials for a single mechanism to a copy of the current
        set, instead of creating a new set for multiple mechanisms.
        Unlike :meth:`acquire`, you cannot pass None desired name or
        mechanism.

        If the `impersonator` argument is used, the credentials will
        impersonate the given name using the impersonator credentials
        (:requires-ext:`s4u`).

        If the `store` argument is used, the credentials will be acquired
        from the given credential store (:requires-ext:`cred_store`).
        Otherwise, the credentials are acquired from the default store.

        The credential store information is a dictionary containing
        mechanisms-specific keys and values pointing to a credential store
        or stores.

        Note that the `store` argument is not compatible with the
        `impersonator` argument.

        Args:
            name (~gssapi.names.Name): the name associated with the
                credentials
            mech (~gssapi.OID): the desired :class:`MechType` to be used with
                the credentials
            usage (str): the usage for the credentials -- either 'both',
                'initiate', or 'accept'
            init_lifetime (int): the desired initiate lifetime of the
                credentials in seconds, or None for indefinite
            accept_lifetime (int): the desired accept lifetime of the
                credentials in seconds, or None for indefinite
            impersonator (Credentials): the credentials to use to impersonate
                the given name, or None to not acquire normally
                (:requires-ext:`s4u`)
            store (dict): the credential store information pointing to the
                credential store from which to acquire the credentials,
                or None for the default store (:requires-ext:`cred_store`)

        Returns:
            Credentials: the credentials set containing the current credentials
            and the newly acquired ones.

        Raises:
            ~gssapi.exceptions.BadMechanismError
            ~gssapi.exceptions.BadNameTypeError
            ~gssapi.exceptions.BadNameError
            ~gssapi.exceptions.DuplicateCredentialsElementError
            ~gssapi.exceptions.ExpiredCredentialsError
            ~gssapi.exceptions.MissingCredentialsError
        """

        if store is not None and impersonator is not None:
            raise ValueError('You cannot use both the `impersonator` and '
                             '`store` arguments at the same time')

        if store is not None:
            if rcred_cred_store is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not have support for manipulating "
                                          "credential stores")
            b_store = _encode_dict(store)

            res = rcred_cred_store.add_cred_from(b_store, self, name, mech,
                                                 usage, init_lifetime,
                                                 accept_lifetime)
        elif impersonator is not None:
            if rcred_s4u is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not have support for S4U")
            res = rcred_s4u.add_cred_impersonate_name(self, impersonator,
                                                      name, mech, usage,
                                                      init_lifetime,
                                                      accept_lifetime)
        else:
            res = rcreds.add_cred(self, name, mech, usage, init_lifetime,
                                  accept_lifetime)

        return Credentials(res.creds)

    def export(self) -> bytes:
        """Export these credentials into a token

        This method exports the current credentials to a token that can
        then be imported by passing the `token` argument to the constructor.

        This is often used to pass credentials between processes.

        :requires-ext:`cred_imp_exp`

        Returns:
            bytes: the exported credentials in token form
        """

        if rcred_imp_exp is None:
            raise NotImplementedError("Your GSSAPI implementation does not "
                                      "have support for importing and "
                                      "exporting creditials")

        return rcred_imp_exp.export_cred(self)

    # pickle protocol support
    def __reduce__(
        self,
    ) -> t.Tuple[t.Type["Credentials"], t.Tuple[None, bytes]]:
        # the unpickle arguments to new are (base=None, token=self.export())
        return (type(self), (None, self.export()))



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.chan_bindings import ChannelBindings
    from gssapi.raw.creds import Creds
    from gssapi.raw.named_tuples import (
        AcceptSecContextResult,
        InitSecContextResult,
        InquireContextResult,
    )
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID
    from gssapi.raw.types import RequirementFlag

class SecurityContext:
    """
    A GSSAPI Security Context
    """

    def __new__(
        cls,
        cpy: t.Optional["SecurityContext"] = None,
    ) -> "SecurityContext": ...

    @property
    def _started(self) -> bool: ...


def init_sec_context(
    name: "Name",
    creds: t.Optional["Creds"] = None,
    context: t.Optional[SecurityContext] = None,
    mech: t.Optional["OID"] = None,
    flags: t.Optional[t.Union[
        int, "RequirementFlag",
        t.Iterable[int], t.Iterable["RequirementFlag"]
    ]] = None,
    lifetime: t.Optional[int] = None,
    channel_bindings: t.Optional["ChannelBindings"] = None,
    input_token: t.Optional[bytes] = None,
) -> "InitSecContextResult":
    """Initiate a GSSAPI security context.

    This method initiates a GSSAPI security context, targeting the given
    target name.  To create a basic context, just provide the target name.
    Further calls used to update the context should pass in the output context
    of the last call, as well as the input token received from the acceptor.

    Warning:
        This changes the input context!

    Args:
        target_name (~gssapi.raw.names.Name): the target for the security
            context
        creds (Creds): the credentials to use to initiate the context,
            or None to use the default credentials
        context (~gssapi.raw.sec_contexts.SecurityContext): the security
            context to update, or None to create a new context
        mech (~gssapi.raw.types.MechType): the mechanism type for this security
            context, or None for the default mechanism type
        flags (list): the flags to request for the security context, or
            None to use the default set: mutual_authentication and
            out_of_sequence_detection.  This may also be an
            :class:`IntEnumFlagSet`
        lifetime (int): the request lifetime of the security context in seconds
            (a value of 0 or None means indefinite)
        channel_bindings (ChannelBindings): The channel bindings (or None for
            no channel bindings)
        input_token (bytes): the token to use to update the security context,
            or None if you are creating a new context

    Returns:
        InitSecContextResult: the output security context, the actual mech
        type, the actual flags used, the output token to send to the acceptor,
        the actual lifetime of the context in seconds (or None if not supported
        or indefinite), and whether or not more calls are needed to finish the
        initiation.

    Raises:
        ~gssapi.exceptions.InvalidTokenError
        ~gssapi.exceptions.InvalidCredentialsError
        ~gssapi.exceptions.MissingCredentialsError
        ~gssapi.exceptions.ExpiredCredentialsError
        ~gssapi.exceptions.BadChannelBindingsError
        ~gssapi.exceptions.BadMICError
        ~gssapi.exceptions.ExpiredTokenError
        ~gssapi.exceptions.DuplicateTokenError
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
        ~gssapi.exceptions.BadMechanismError
    """


def accept_sec_context(
    input_token: bytes,
    acceptor_creds: t.Optional["Creds"] = None,
    context: t.Optional[SecurityContext] = None,
    channel_bindings: t.Optional["ChannelBindings"] = None,
) -> "AcceptSecContextResult":
    """Accept a GSSAPI security context.

    This method accepts a GSSAPI security context using a token sent by the
    initiator, using the given credentials.  It can either be used to accept a
    security context and create a new security context object, or to update an
    existing security context object.

    Warning:
        This changes the input context!

    Args:
        input_token (bytes): the token sent by the context initiator
        acceptor_creds (Creds): the credentials to be used to accept the
            context (or None to use the default credentials)
        context (~gssapi.raw.sec_contexts.SecurityContext): the security
            context to update (or None to create a new security context object)
        channel_bindings (ChannelBindings): The channel bindings (or None for
            no channel bindings)

    Returns:
        AcceptSecContextResult: the resulting security context, the initiator
        name, the mechanism being used, the output token, the flags in use,
        the lifetime of the context in seconds (or None for indefinite or not
        supported), the delegated credentials (valid only if the
        delegate_to_peer flag is set), and whether or not further token
        exchanges are needed to finalize the security context.

    Raises:
        ~gssapi.exceptions.InvalidTokenError
        ~gssapi.exceptions.InvalidCredentialsError
        ~gssapi.exceptions.MissingCredentialsError
        ~gssapi.exceptions.ExpiredCredentialsError
        ~gssapi.exceptions.BadChannelBindingsError
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.BadMICError
        ~gssapi.exceptions.ExpiredTokenError
        ~gssapi.exceptions.DuplicateTokenError
        ~gssapi.exceptions.BadMechanismError
    """


def inquire_context(
    context: SecurityContext,
    initiator_name: bool = True,
    target_name: bool = True,
    lifetime: bool = True,
    mech: bool = True,
    flags: bool = True,
    locally_init: bool = True,
    complete: bool = True,
) -> "InquireContextResult":
    """Get information about a security context.

    This method obtains information about a security context, including
    the initiator and target names, as well as the TTL, mech,
    flags, and its current state (open vs closed).

    Note:
        the target name may be ``None`` if it would have been ``GSS_C_NO_NAME``

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the context in
            question

    Returns:
        InquireContextResult: the initiator name, the target name, the TTL
        (can be None for indefinite or not supported), the mech type, the
        flags, whether or not the context was locally initiated,
        and whether or not the context is currently fully established

    Raises:
        ~gssapi.exceptions.MissingContextError
    """


def context_time(
    context: SecurityContext,
) -> int:
    """Get the amount of time for which the given context will remain valid.

    This method determines the amount of time for which the given
    security context will remain valid.  An expired context will
    give a result of 0.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the security
            context in question

    Returns:
        int: the number of seconds for which the context will be valid

    Raises:
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.MissingContextError
    """


def process_context_token(
    context: SecurityContext,
    token: bytes,
) -> None:
    """Process a token asynchronously.

    This method provides a way to process a token, even if the
    given security context is not expecting one.  For example,
    if the initiator has the initSecContext return that the context
    is complete, but the acceptor is unable to accept the context,
    and wishes to send a token to the initiator, letting the
    initiator know of the error.

    Warning:
        This method has been essentially deprecated by :rfc:`2744`.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the security
            context against which to process the token
        token (bytes): the token to process

    Raises:
        ~gssapi.exceptions.InvalidTokenError
        ~gssapi.exceptions.MissingContextError
    """


def import_sec_context(
    token: bytes,
) -> SecurityContext:
    """Import a context from another process.

    This method imports a security context established in another process
    by reading the specified token which was output by
    :func:`export_sec_context`.

    Raises:
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.InvalidTokenError
        ~gssapi.exceptions.OperationUnavailableError
        ~gssapi.exceptions.UnauthorizedError
    """


def export_sec_context(
    context: SecurityContext,
) -> bytes:
    """Export a context for use in another process.

    This method exports a security context, deactivating in the current process
    and creating a token which can then be imported into another process
    with :func:`import_sec_context`.

    Warning: this modifies the input context

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the context to send
            to another process

    Returns:
        bytes: the output token to be imported

    Raises:
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.OperationUnavailableError
    """


def delete_sec_context(
    context: SecurityContext,
    local_only: bool = True,
) -> bytes:
    """Delete a GSSAPI security context.

    This method deletes a GSSAPI security context,
    returning an output token to send to the other
    holder of the security context to notify them
    of the deletion.

    Note:
        This method generally should not be used.  :class:`SecurityContext`
        objects will automatically be freed by Python.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the security
            context in question
        local_only (bool): should we request local deletion (True), or also
            remote deletion (False), in which case a token is also returned

    Returns:
        bytes: the output token (if remote deletion is requested).  Generally
            this is None, but bytes for compatibility.

    Raises:
        ~gssapi.exceptions.MissingContextError
    """



"""Credential Store Extension"""
import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.named_tuples import AcquireCredResult, StoreCredResult
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID


def acquire_cred_from(
    dict_store: t.Optional[
        t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
    ] = None,
    name: t.Optional["Name"] = None,
    lifetime: t.Optional[int] = None,
    mechs: t.Optional[t.Iterable["OID"]] = None,
    usage: str = 'both',
) -> "AcquireCredResult":
    """Acquire credentials from the given store.

    This method acquires credentials from the store specified by the
    given credential store information.

    The credential store information is a dictionary containing
    mechanisms-specific keys and values pointing to a credential store
    or stores.

    Args:
        store (dict): the credential store information pointing to the
            credential store from which to acquire the credentials.
            See :doc:`credstore` for valid values
        name (~gssapi.raw.names.Name): the name associated with the
            credentials, or None for the default name
        lifetime (int): the desired lifetime of the credentials in seconds, or
            None for indefinite
        mechs (list): the desired mechanisms to be used with these
            credentials, or None for the default set
        usage (str): the usage for these credentials -- either 'both',
            'initiate', or 'accept'

    Returns:
        AcquireCredResult: the acquired credentials and information about
        them

    Raises:
        ~gssapi.exceptions.GSSError
    """


def add_cred_from(
    dict_store: t.Optional[
        t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
    ],
    input_creds: "Creds",
    name: "Name",
    mech: "OID",
    usage: str = 'both',
    init_lifetime: t.Optional[int] = None,
    accept_lifetime: t.Optional[int] = None,
) -> "AcquireCredResult":
    """Acquire credentials to add to the current set from the given store.

    This method works like :func:`acquire_cred_from`, except that it
    adds the acquired credentials for a single mechanism to a copy of
    the current set, instead of creating a new set for multiple mechanisms.
    Unlike :func:`~gssapi.raw.creds.acquire_cred`, you cannot pass None for the
    desired name or mechanism.

    The credential store information is a dictionary containing
    mechanisms-specific keys and values pointing to a credential store
    or stores.

    Args:
        store (dict): the store into which to store the credentials,
            or None for the default store.
            See :doc:`credstore` for valid values
        name (~gssapi.raw.names.Name): the name associated with the credentials
        mech (~gssapi.OID): the desired mechanism to be used with these
            credentials
        usage (str): the usage for these credentials -- either 'both',
            'initiate', or 'accept'
        init_lifetime (int): the desired initiate lifetime of the credentials
            in seconds, or None for indefinite
        accept_lifetime (int): the desired accept lifetime of the credentials
            in seconds, or None for indefinite

    Returns:
        AcquireCredResult: the new credentials set and information about
        it

    Raises:
        ~gssapi.exceptions.GSSError
    """


def store_cred_into(
    dict_store: t.Optional[
        t.Dict[t.Union[bytes, str], t.Union[bytes, str]]
    ],
    creds: "Creds",
    usage: str = 'both',
    mech: t.Optional["OID"] = None,
    overwrite: bool = False,
    set_default: bool = False,
) -> "StoreCredResult":
    """Store credentials into the given store.

    This method stores the given credentials into the store specified
    by the given store information.  They may then be retrieved later using
    :func:`acquire_cred_from` or :func:`add_cred_from`.

    The credential store information is a dictionary containing
    mechanisms-specific keys and values pointing to a credential store
    or stores.

    Args:
        store (dict): the store into which to store the credentials,
            or None for the default store.
            See :doc:`credstore` for valid values
        creds (Creds): the credentials to store
        usage (str): the usage to store the credentials with -- either
            'both', 'initiate', or 'accept'
        mech (~gssapi.OID): the mechansim to associate with the stored
            credentials
        overwrite (bool): whether or not to overwrite existing credentials
            stored with the same name, etc
        set_default (bool): whether or not to set these credentials as
            the default credentials for the given store.

    Returns:
        StoreCredResult: the results of the credential storing operation

    Raises:
        ~gssapi.exceptions.GSSError
    """



from gssapi.raw.misc import GSSError

class ParameterReadError(GSSError): ...

class ParameterWriteError(GSSError): ...

class MalformedParameterError(GSSError): ...

class BadMechanismError(GSSError): ...

class BadNameError(GSSError): ...

class BadNameTypeError(GSSError): ...

class BadChannelBindingsError(GSSError): ...

class BadStatusError(GSSError): ...

class BadMICError(GSSError): ...

class MissingCredentialsError(GSSError): ...

class MissingContextError(GSSError): ...

class InvalidTokenError(GSSError): ...

class InvalidCredentialsError(GSSError): ...

class ExpiredCredentialsError(GSSError): ...

class ExpiredContextError(GSSError): ...

class BadQoPError(GSSError): ...

class UnauthorizedError(GSSError): ...

class OperationUnavailableError(GSSError): ...

class DuplicateCredentialsElementError(GSSError): ...

class MechanismNameRequiredError(GSSError): ...

class NameReadError(ParameterReadError, BadNameError): ...

class NameTypeReadError(ParameterReadError, BadNameTypeError): ...

class TokenReadError(ParameterReadError, InvalidTokenError): ...

class ContextReadError(ParameterReadError, MissingContextError): ...

class CredentialsReadError(ParameterReadError, MissingCredentialsError): ...

class ContextWriteError(ParameterWriteError, MissingContextError): ...

class CredentialsWriteError(ParameterWriteError, MissingCredentialsError): ...

class SupplementaryError(GSSError): ...

class DuplicateTokenError(SupplementaryError): ...

class ExpiredTokenError(SupplementaryError): ...

class TokenOutOfSequenceError(SupplementaryError): ...

class TokenTooLateError(TokenOutOfSequenceError): ...

class TokenTooEarlyError(TokenOutOfSequenceError): ...



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import InquireAttrsResult, DisplayAttrResult
    from gssapi.raw.oids import OID


def indicate_mechs_by_attrs(
    desired_mech_attrs: t.Optional[t.Iterable["OID"]] = None,
    except_mech_attrs: t.Optional[t.Iterable["OID"]] = None,
    critical_mech_attrs: t.Optional[t.Iterable["OID"]] = None,
) -> t.Set["OID"]:
    """Get a set of mechanisms that have the specified attributes.

    Args:
        desired_mech_attrs (~gssapi.OID): Attributes that the output mechs MUST
            offer
        except_mech_attrs (~gssapi.OID): Attributes that the output mechs MUST
            NOT offer
        critical_mech_attrs (~gssapi.OID): Attributes that the output mechs
            MUST understand and offer

    Returns:
        ~gssapi.MechType: a set of mechs which satisfy the given criteria

    Raises:
        ~gssapi.exceptions.GSSError
    """


def inquire_attrs_for_mech(
    mech: "OID",
) -> "InquireAttrsResult":
    """Gets the set of attrs supported and known by a mechanism.

    Args:
        mech (~gssapi.raw.types.MechType): Mechanism to inquire about

    Returns:
        InquireAttrsResult: the results of inquiry; a mech's attributes and
        known attributes

    Raises:
        ~gssapi.exceptions.GSSError
    """


def display_mech_attr(
    attr: "OID",
) -> "DisplayAttrResult":
    """Returns information about attributes in human readable form.

    Args:
        attr (~gssapi.OID): Mechanism attribute to retrieve names and
            descriptions of

    Returns:
        DisplayAttrResult: the results of displaying the attribute; mech name,
        short description, and long description.

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.named_tuples import StoreCredResult
    from gssapi.raw.oids import OID


def store_cred(
    creds: "Creds",
    usage: str = 'both',
    mech: t.Optional["OID"] = None,
    overwrite: bool = False,
    set_default: bool = False,
) -> "StoreCredResult":
    """Store credentials into the default store.

    This method stores the given credentials into the default store.
    They may then be retrieved later using
    :func:`~gssapi.raw.creds.acquire_cred`.

    Args:
        creds (Creds): the credentials to store
        usage (str): the usage to store the credentials with -- either
            'both', 'initiate', or 'accept'
        mech (~gssapi.OID): the mechansim to associate with the stored
            credentials
        overwrite (bool): whether or not to overwrite existing credentials
            stored with the same name, etc
        set_default (bool): whether or not to set these credentials as
            the default credentials for the given store.

    Returns:
        StoreCredResult: the results of the credential storing operation

    Raises:
        ~gssapi.exceptions.GSSError
        ~gssapi.exceptions.ExpiredCredentialsError
        ~gssapi.exceptions.MissingCredentialsError
        ~gssapi.exceptions.OperationUnavailableError
        ~gssapi.exceptions.DuplicateCredentialsElementError
    """



"""Low-Level GSSAPI Bindings

The low-level API presents a series of methods designed
to closely mimic the C API presented in RFC 2744 and
associated RFCs.

In this API, classes are simply thin wrappers around C
constructs, and generally lack instance methods.  However,
classes will automatically free associated memory (so the
release_xyz methods are not necessary to call).

The core RFC 2744 components are organized into the following
submodules:

    gssapi.raw.names -- Names

    gssapi.raw.creds -- Credentials

    gssapi.raw.sec_contexts -- Security Contexts

    gssapi.raw.message -- Message encryption, decryption, etc

    gssapi.raw.misc -- Miscellaneous functions

    gssapi.raw.types -- Miscellaneous types (enums, etc)

    gssapi.raw.exceptions -- Exceptions

Additionally, a number of extensions may be present.  All extensions
are in modules of the form `gssapi.raw.ext_xyz`.

All available functions and classes can be accessed directly from this
module (`gssapi.raw`) -- it is unnecessary to directly import submodules.
"""


import pkgutil
import importlib

from gssapi.raw import _enum_extensions

# NB(directxman12): the enum extensions must be imported BEFORE ANYTHING ELSE!
for modinf in pkgutil.iter_modules(_enum_extensions.__path__):
    name = modinf[1]
    importlib.import_module('{0}._enum_extensions.{1}'.format(__name__, name))

del pkgutil
del importlib

from gssapi.raw.creds import *  # noqa
from gssapi.raw.message import *  # noqa
from gssapi.raw.misc import *  # noqa
from gssapi.raw.exceptions import *  # noqa
from gssapi.raw.names import *  # noqa
from gssapi.raw.sec_contexts import *  # noqa
from gssapi.raw.oids import *  # noqa
from gssapi.raw.types import *  # noqa
from gssapi.raw.chan_bindings import *  # noqa

# optional S4U support
try:
    from gssapi.raw.ext_s4u import *  # noqa
except ImportError:
    pass  # no s4u support in the system's GSSAPI library

# optional cred store support
try:
    from gssapi.raw.ext_cred_store import *  # noqa
except ImportError:
    pass

# optional RFC 4178 support
try:
    from gssapi.raw.ext_rfc4178 import *  # noqa
except ImportError:
    pass

# optional RFC 5587 support
try:
    from gssapi.raw.ext_rfc5587 import *  # noqa
except ImportError:
    pass

# optional RFC 5588 support
try:
    from gssapi.raw.ext_rfc5588 import *  # noqa
except ImportError:
    pass

# optional RFC 5801 support
try:
    from gssapi.raw.ext_rfc5801 import *  # noqa
except ImportError:
    pass

try:
    from gssapi.raw.ext_cred_imp_exp import *  # noqa
except ImportError:
    pass

# optional KRB5 mech support
try:
    import gssapi.raw.mech_krb5  # noqa
except ImportError:
    pass

# optional password support
try:
    from gssapi.raw.ext_password import *  # noqa
    from gssapi.raw.ext_password_add import *  # noqa
except ImportError:
    pass

# optional DCE (IOV) support
try:
    from gssapi.raw.ext_dce import *  # noqa
    # optional IOV MIC support (requires DCE support)
    from gssapi.raw.ext_iov_mic import *  # noqa
except ImportError:
    pass

# optional DCE (AEAD) support
try:
    from gssapi.raw.ext_dce_aead import *  # noqa
except ImportError:
    pass

# optional KRB5 specific extension support
try:
    from gssapi.raw.ext_krb5 import *  # noqa
except ImportError:
    pass

# optional RFC 6680 support
try:
    from gssapi.raw.ext_rfc6680 import *  # noqa
    from gssapi.raw.ext_rfc6680_comp_oid import *  # noqa
except ImportError:
    pass

# optional Global Grid Forum support
try:
    from gssapi.raw.ext_ggf import *  # noqa
except ImportError:
    pass

# optional set_cred_option support
try:
    from gssapi.raw.ext_set_cred_opt import *  # noqa
except ImportError:
    pass



import typing as t

from enum import IntEnum

from gssapi.raw.ext_dce_aead import wrap_aead, unwrap_aead

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import IOVUnwrapResult, WrapResult
    from gssapi.raw.sec_contexts import SecurityContext


class IOVBufferType(IntEnum):
    """
    IOV Buffer Types

    This IntEnum represent GSSAPI IOV buffer
    types to be used with the IOV methods.

    The numbers behind the values correspond directly
    to their C counterparts.
    """

    empty = 0 #: GSS_IOV_BUFFER_TYPE_EMPTY
    data = 1 #: GSS_IOV_BUFFER_TYPE_DATA
    header = 2 #: GSS_IOV_BUFFER_TYPE_HEADER
    mech_params = 3 #: GSS_IOV_BUFFER_TYPE_MECH_PARAMS
    trailer = 7 #: GSS_IOV_BUFFER_TYPE_TRAILER
    padding = 9 #: GSS_IOV_BUFFER_TYPE_PADDING
    stream = 10 #: GSS_IOV_BUFFER_TYPE_STREAM
    sign_only = 11 #: GSS_IOV_BUFFER_TYPE_SIGN_ONLY
    mic_token = 12 #: GSS_IOV_BUFFER_TYPE_MIC_TOKEN


class IOVBuffer(t.NamedTuple):
    type: IOVBufferType
    allocate: t.Optional[bool]
    value: t.Optional[bytes]


class IOV:
    """A GSSAPI IOV"""

    def __init__(
        self,
        *args: t.Union[
            IOVBuffer,
            t.Tuple[
                t.Union[IOVBufferType, int],
                t.Optional[bool],
                t.Optional[bytes]],
            t.Tuple[
                t.Union[IOVBufferType, int],
                t.Optional[t.Union[bool, bytes]],
            ],
            bytes,
            t.Union[IOVBufferType, int],
        ],
        std_layout: bool = True,
        auto_alloc: bool = True,
    ) -> None: ...

    def __getitem__(
        self,
        ind: int,
    ) -> IOVBuffer: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> t.Iterator[IOVBuffer]: ...

    def __contains__(
        self,
        item: IOVBuffer,
    ) -> bool: ...

    def __reversed__(self) -> t.Iterator[IOVBuffer]: ...

    def index(
        self,
        value: t.Any,
    ) -> int: ...

    def count(
        self,
        value: t.Any,
    ) -> int: ...


def wrap_iov(
    context: "SecurityContext",
    message: IOV,
    confidential: bool = True,
    qop: t.Optional[int] = None,
) -> bool:
    """Wrap/Encrypt an IOV message.

    This method wraps or encrypts an IOV message.  The allocate
    parameter of the :class:`IOVBuffer` objects in the :class:`IOV`
    indicates whether or not that particular buffer should be
    automatically allocated (for use with padding, header, and
    trailer buffers).

    Warning:
        This modifies the input :class:`IOV`.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (IOV): an :class:`IOV` containing the message
        confidential (bool): whether or not to encrypt the miovessage (True),
            or just wrap it with a MIC (False)
        qop (int): the desired Quality of Protection
            (or None for the default QoP)

    Returns:
        bool: whether or not confidentiality was actually used

    Raises:
        ~gssapi.exceptions.GSSError
    """


def unwrap_iov(
    context: "SecurityContext",
    message: IOV,
) -> "IOVUnwrapResult":
    """Unwrap/Decrypt an IOV message.

    This method uwraps or decrypts an IOV message.  The allocate
    parameter of the :class:`IOVBuffer` objects in the :class:`IOV`
    indicates whether or not that particular buffer should be
    automatically allocated (for use with padding, header, and
    trailer buffers).

    As a special case, you may pass an entire IOV message
    as a single 'stream'.  In this case, pass a buffer type
    of :attr:`IOVBufferType.stream` followed by a buffer type of
    :attr:`IOVBufferType.data`.  The former should contain the
    entire IOV message, while the latter should be empty.

    Warning:
        This modifies the input :class:`IOV`.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (IOV): an :class:`IOV` containing the message

    Returns:
        IOVUnwrapResult: whether or not confidentiality was used,
        and the QoP used.

    Raises:
        ~gssapi.exceptions.GSSError
    """


def wrap_iov_length(
    context: "SecurityContext",
    message: IOV,
    confidential: bool = True,
    qop: t.Optional[int] = None,
) -> "WrapResult":
    """Appropriately size padding, trailer, and header IOV buffers.

    This method sets the length values on the IOV buffers.  You
    should already have data provided for the data (and sign-only)
    buffer(s) so that padding lengths can be appropriately computed.

    In Python terms, this will result in an appropriately sized
    `bytes` object consisting of all zeros.

    Warning:
        This modifies the input :class:`IOV`.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (IOV): an :class:`IOV` containing the message

    Returns:
        WrapResult: a list of :class:IOVBuffer` objects, and whether or not
        encryption was actually used

    Raises:
        ~gssapi.exceptions.GSSError
    """



"""
gss_set_cred_option

Provides a way to set options on a credential based on the OID specified. A
common use case is to set the GSS_KRB5_CRED_NO_CI_FLAGS_X on a Kerberos
credential. This is used for interoperability with Microsoft's SSPI.

Note this function is commonly lumped with the GGF extensions but they are not
part of the GGF IETF draft so it's separated into it's own file.

Closest draft IETF document for the gss_set_cred_option can be found at
https://tools.ietf.org/html/draft-williams-kitten-channel-bound-flag-01
"""
import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.oids import OID


def set_cred_option(
    desired_aspect: "OID",
    creds: t.Optional["Creds"] = None,
    value: t.Optional[bytes] = None,
) -> "Creds":
    """
    This method is used to set options of a :class:`~gssapi.raw.creds.Creds`
    object based on an OID key. The options that can be set depends on the mech
    the credentials were created with.

    An example of how this can be used would be to set the
    GSS_KRB5_CRED_NO_CI_FLAGS_X on a Kerberos credential. The OID string for
    this flag is '1.2.752.43.13.29' and it requires no value to be set. This
    must be set before the SecurityContext was initialised with the
    credentials.

    Args:
        desired_aspect (~gssapi.raw.oids.OID): the desired aspect of the
            Credential to set.
        cred_handle (~gssapi.raw.creds.Creds): the Credentials to set, or None
            to create a new credential.
        value (bytes): the value to set on the desired aspect of the Credential
            or None to send GSS_C_EMPTY_BUFFER.

    Returns:
        Creds: The output credential.

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

from gssapi.raw.exceptions import *  # noqa
from gssapi.raw.misc import GSSError  # noqa

"""High-Level API Errors

This module includes several high-level exceptions,
in addition to GSSError and exceptions from
:mod:`gssapi.raw.exceptions`.
"""


# non-GSS exceptions
class GeneralError(Exception):
    """A General High-Level API Error"""
    MAJOR_MESSAGE = "General error"
    FMT_STR = "{maj}: {min}."

    def __init__(
        self,
        minor_message: str,
        **kwargs: str,
    ) -> None:
        maj_str = self.MAJOR_MESSAGE.format(**kwargs)
        err_str = self.FMT_STR.format(maj=maj_str, min=minor_message)
        super(GeneralError, self).__init__(err_str)


class UnknownUsageError(GeneralError):
    """An Error indicating an unknown usage type"""
    MAJOR_MESSAGE = "Unable to determine {obj} usage"


class EncryptionNotUsed(GeneralError):
    """An Error indicating that encryption was requested, but not used"""
    MAJOR_MESSAGE = "Confidentiality was requested, but not used"

    def __init__(
        self,
        minor_message: str,
        unwrapped_message: t.Optional[bytes] = None,
        **kwargs: str,
    ) -> None:
        super(EncryptionNotUsed, self).__init__(minor_message, **kwargs)

        self.unwrapped_message = unwrapped_message



import typing as t

from gssapi.raw.names import Name
from gssapi.raw.oids import OID
from gssapi.raw.types import MechType


def indicate_mechs() -> t.Set[OID]:
    """Get the currently supported mechanisms.

    This method retrieves the currently supported GSSAPI mechanisms.
    Note that if unknown mechanims are found, those will be skipped.
    """


def inquire_names_for_mech(
    mech: OID,
) -> t.Set[OID]:
    """
    inquire_names_for_mech(mech)
    Get the name types supported by a mechanism.

    This method retrieves the different name types supported by
    the given mechanism.

    Args:
        mech (~gssapi.OID): the mechanism in question

    Returns:
        list: the name type OIDs supported by the given mechanism

    Raises:
        ~gssapi.exceptions.GSSError
    """


def inquire_mechs_for_name(
    name: Name,
) -> t.Set[OID]:
    """
    inquire_mechs_for_name(name)
    List the mechanisms which can process a name.

    This method lists the mechanisms which may be able to
    process the given name.

    Args:
        name (~gssapi.raw.names.Name): the name in question

    Returns:
        The mechanism OIDs able to process the given name

    Raises:
        ~gssapi.exceptions.GSSError
    """


def _display_status(
    error_code: int,
    is_major_code: bool,
    mech: t.Optional[MechType] = None,
    message_context: int = 0,
) -> t.Tuple[bytes, int, bool]:
    """
    Display a string message for a GSSAPI error code.

    This method displays a message for a corresponding GSSAPI error code.
    Since some error codes might have multiple messages, a context parameter
    may be passed to indicate where in the series of messages we currently are
    (this is the second item in the return value tuple).  Additionally, the
    third item in the return value tuple indicates whether or not more
    messages are available.

    Args:
        error_code (int): The error code in question
        is_major_code (bool): is this a major code (True) or a
            minor code (False)
        mech (~gssapi.raw.types.MechType): The mechanism type that returned
            this error code (defaults to None, for the default mechanism)
        message_context (int): The context for this call -- this is used when
            multiple messages are available (defaults to 0)

    Returns:
        (bytes, int, bool): the message, the new message context, and
            whether or not to call again for further messages

    Raises:
       ValueError
    """


class GSSError(Exception):
    """
    A GSSAPI Error

    This Exception represents an error returned from the GSSAPI
    C bindings.  It contains the major and minor status codes
    returned by the method which caused the error, and can
    generate human-readable string messages from the error
    codes
    """

    maj_code: int
    min_code: int
    token: t.Optional[bytes]
    calling_code: int
    routine_code: int
    supplementary_code: int

    @classmethod
    def _parse_major_code(
        cls,
        maj_code: int
    ) -> t.Tuple[int, int, int]: ...

    def __init__(
        self,
        maj_code: int,
        min_code: int,
        token: t.Optional[bytes] = None,
    ) -> None:
        """
        Create a new GSSError.

        This method creates a new GSSError,
        retrieves the related human-readable
        string messages, and uses the results to construct an
        exception message

        Args:
            maj_code: the major code associated with this error
            min_code: the minor code associated with this error
            token: an error token associated with the error
        """

    def get_all_statuses(
        self,
        code: int,
        is_maj: bool,
    ) -> t.List[str]:
        """
        Retrieve all messages for a status code.

        This method retrieves all human-readable messages
        available for the given status code.

        Args:
            code: the status code in question
            is_maj: whether this is a major status code (True)
                or minor status code (False)

        Returns:
            [str]: A list of string messages associated with the
                given code
        """

    def gen_message(self) -> str:
        """
        Retrieves all messages for this error's status codes

        This method retrieves all messages for this error's status codes,
        and forms them into a string for use as an exception message

        Returns:
            str: a string for use as this error's message
        """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.named_tuples import AddCredResult
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID


def add_cred_with_password(
    input_cred: "Creds",
    name: "Name",
    mech: "OID",
    password: bytes,
    usage: str = 'initiate',
    init_lifetime: t.Optional[int] = None,
    accept_lifetime: t.Optional[int] = None,
) -> "AddCredResult":
    """Add a credential-element to a credential using provided password.

    This function is originally from Solaris and is not documented by either
    MIT or Heimdal.

    In general, it functions similarly to :func:`~gssapi.raw.creds.add_cred`.

    Args:
        input_cred (~gssapi.raw.creds.Creds): the credentials to add to
        name (~gssapi.raw.names.Name): the name to acquire credentials for
        mech (~gssapi.raw.types.MechType): the desired mechanism.  Note that
            this is both singular and required
        password (bytes): the password used to acquire credentialss with
        usage (str): the usage type for the credentials: may be
            'initiate', 'accept', or 'both'
        init_lifetime (int): the lifetime, in seconds, for the credentials to
            remain valid when using them to initiate security contexts (or None
            for indefinite)
        accept_lifetime (int): the lifetime, in seconds, for the credentials to
            remain valid when using them to accept security contexts (or None
            for indefinite)

    Returns:
        AddCredResult: the actual mechanisms with which the credentials may be
        used, the actual initiator TTL in seconds, and the actual acceptor TTL
        in seconds (the TTLs may be None for indefinite or not supported)

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import InquireSASLNameResult
    from gssapi.raw.oids import OID


def inquire_saslname_for_mech(
    mech: "OID",
) -> "InquireSASLNameResult":
    """Gets information about a specified mech, including the SASL name,
    the mech name, and the mech description.

    Args:
        mech (~gssapi.OID): Mechanism to inquire about

    Returns:
        InquireSASLNameResult: the results of inquiry; a mech's SASL name,
        name, and description.

    Raises:
        ~gssapi.exceptions.GSSError: an unknown failure occurred
    """


def inquire_mech_for_saslname(
    sasl_name: bytes,
) -> "OID":
    """Gets the OID for the mech specified by SASL name.

    Args:
        sasl_name (bytes): SASL name of the mechanism

    Returns:
        ~gssapi.OID: the mechanism with corresponding SASL name.

    Raises:
        ~gssapi.exceptions.GSSError: An unknown failure occurred
    """



"""Credentials Import/Export Extension"""

import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds


def export_cred(
    creds: "Creds",
) -> bytes:
    """Export GSSAPI credentials.

    This method exports GSSSAPI credentials into a token
    which may be transmitted between different processes.

    Args:
        creds (Creds): the credentials object to be exported

    Returns:
        bytes: the exported token representing the given credentials object

    Raises:
        ~gssapi.exceptions.GSSError
    """


def import_cred(
    token: bytes,
) -> "Creds":
    """Import GSSAPI credentials from a token.

    This method imports a credentials object from a token
    previously exported by :func:`export_cred`.

    Args:
        token (bytes): the token to import

    Returns:
        Creds: the imported credentials object

    Raises:
        ~gssapi.exceptions.GSSError
    """



"""Service4User Extension"""
import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.named_tuples import AcquireCredResult, AddCredResult
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID


def acquire_cred_impersonate_name(
    impersonator_cred: "Creds",
    name: "Name",
    lifetime: t.Optional[int] = None,
    mechs: t.Optional[t.Iterable["OID"]] = None,
    usage: str = 'initiate',
) -> "AcquireCredResult":
    """Acquire credentials by impersonating another name.

    This method is one of the ways to use S4U2Self.  It acquires credentials
    by impersonating another name using a set of proxy credentials.  The
    impersonator credentials must have a usage of 'both' or 'initiate'.

    Args:
        impersonator_cred (~gssapi.raw.creds.Creds): the credentials with
            permissions to impersonate the target name
        name (~gssapi.raw.names.Name): the name to impersonate
        lifetime (int): the lifetime for the credentials (or None for
            indefinite) in seconds
        mechs (~gssapi.raw.types.MechType): the desired mechanisms for which
            the credentials should work (or None for the default set)
        usage (str): the usage type for the credentials: may be
            'initiate', 'accept', or 'both'

    Returns:
        AcquireCredResult: the resulting credentials, the actual mechanisms
        with which they may be used, and their actual lifetime in seconds (or
        None for indefinite or not support)

    Raises:
        ~gssapi.exceptions.GSSError
    """


def add_cred_impersonate_name(
    input_cred: "Creds",
    impersonator_cred: "Creds",
    name: "Name",
    mech: "OID",
    usage: str = 'initiate',
    init_lifetime: t.Optional[int] = None,
    accept_lifetime: t.Optional[int] = None,
) -> "AddCredResult":
    """Add a credentials element to a credential by impersonating another name.

    This method is one of the ways to use S4U2Self.  It adds credentials
    to the input credentials by impersonating another name using a set of
    proxy credentials.  The impersonator credentials must have a usage of
    'both' or 'initiate'.

    Args:
        input_cred (~gssapi.raw.creds.Creds): the set of credentials to which
            to add the new credentials
        impersonator_cred (~gssapi.raw.creds.Creds): the credentials with
            permissions to impersonate the target name
        name (~gssapi.raw.names.Name): the name to impersonate
        mech (~gssapi.raw.types.MechType): the desired mechanism. Note that
            this is both
            singular and required, unlike acquireCredImpersonateName
        usage (str): the usage type for the credentials: may be
            'initiate', 'accept', or 'both'
        init_lifetime (int): the lifetime, in seconds, for the credentials to
            remain valid when using them to initiate security contexts (or None
            for indefinite)
        accept_lifetime (int): the lifetime, in seconds, for the credentials to
            remain valid when using them to accept security contexts (or None
            for indefinite)

    Returns:
        AddCredResult: the actual mechanisms with which the credentials may be
        used, the actual initiator TTL in seconds, and the actual acceptor TTL
        in seconds (the TTLs may be None for indefinite or not supported)

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.ext_dce import IOV
    from gssapi.raw.sec_contexts import SecurityContext


def get_mic_iov(
    context: "SecurityContext",
    message: "IOV",
    qop: t.Optional[int] = None,
) -> None:
    """Generate MIC tokens for the given IOV message.

    This method generates a MIC token for the given IOV message, and places it
    in the :attr:`~gssapi.raw.ext_dce.IOVBufferType.mic_token` buffer in the
    IOV. This method operates entirely in-place, and returns nothing.

    Warning:
        This modifies the input :class:`~gssapi.raw.ext_dce.IOV`.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (IOV): the :class:`~gssapi.raw.ext_dce.IOV` containing the
            message
        qop (int): the desired Quality of Protection
            (or None for the default QoP)

    Returns:
        None

    Raises:
        ~gssapi.exceptions.GSSError
    """


def get_mic_iov_length(
    context: "SecurityContext",
    message: "IOV",
    qop: t.Optional[int] = None,
) -> None:
    """Allocate space for the MIC buffer in the given IOV message.

    This method allocates space for the MIC token buffer
    (:attr:`~gssapi.raw.ext_dce.IOVBufferType.mic_token`) in the given IOV
    message.

    Warning:
        This modifies the input :class:`~gssapi.raw.ext_dce.IOV`.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (IOV): the :class:`~gssapi.raw.ext_dce.IOV` containing the
            message
        qop (int): the desired Quality of Protection
            (or None for the default QoP)

    Returns:
        None

    Raises:
        ~gssapi.exceptions.GSSError
    """


def verify_mic_iov(
    context: "SecurityContext",
    message: "IOV",
    qop: t.Optional[int] = None,
) -> int:
    """Verify that the MIC matches the data in the given IOV message.

    This method verifies that the MIC token in the MIC buffer
    (:attr:`~gssapi.raw.ext_dce.IOVBufferType.mic_token`) match the data
    buffer(s) in the given IOV method.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (IOV): the :class:`~gssapi.raw.ext_dce.IOV` containing the
            message

    Returns:
        int: the QoP used to generate the MIC token

    Raises:
        ~gssapi.exceptions.GSSError
    """



import sys
import types
import typing as t

import decorator as deco

from gssapi.raw.misc import GSSError

if t.TYPE_CHECKING:
    from gssapi.sec_contexts import SecurityContext


def import_gssapi_extension(
    name: str,
) -> t.Optional[types.ModuleType]:
    """Import a GSSAPI extension module

    This method imports a GSSAPI extension module based
    on the name of the extension (not including the
    'ext_' prefix).  If the extension is not available,
    the method retuns None.

    Args:
        name (str): the name of the extension

    Returns:
        module: Either the extension module or None
    """

    try:
        path = 'gssapi.raw.ext_{0}'.format(name)
        __import__(path)
        return sys.modules[path]
    except ImportError:
        return None


def inquire_property(
    name: str,
    doc: t.Optional[str] = None
) -> property:
    """Creates a property based on an inquire result

    This method creates a property that calls the
    :python:`_inquire` method, and return the value of the
    requested information.

    Args:
        name (str): the name of the 'inquire' result information

    Returns:
        property: the created property
    """

    def inquire_property(self: "SecurityContext") -> t.Any:
        if not self._started:
            msg = (f"Cannot read {name} from a security context whose "
                   "establishment has not yet been started.")
            raise AttributeError(msg)

        return getattr(self._inquire(**{name: True}), name)

    return property(inquire_property, doc=doc)


# use UTF-8 as the default encoding, like Python 3
_ENCODING = 'UTF-8'


def _get_encoding() -> str:
    """Gets the current encoding used for strings.

    This value is used to encode and decode string
    values like names.

    Returns:
        str: the current encoding
    """
    return _ENCODING


def set_encoding(
    enc: str,
) -> None:
    """Sets the current encoding used for strings

    This value is used to encode and decode string
    values like names.

    Args:
        enc: the encoding to use
    """

    global _ENCODING
    _ENCODING = enc


def _encode_dict(
    d: t.Dict[t.Union[bytes, str], t.Union[bytes, str]],
) -> t.Dict[bytes, bytes]:
    """Encodes any relevant strings in a dict"""
    def enc(x: t.Union[bytes, str]) -> bytes:
        if isinstance(x, str):
            return x.encode(_ENCODING)
        else:
            return x

    return {enc(k): enc(v) for k, v in d.items()}


# in case of Python 3, just use exception chaining
@deco.decorator
def catch_and_return_token(
    func: t.Callable,
    self: "SecurityContext",
    *args: t.Any,
    **kwargs: t.Any,
) -> t.Optional[bytes]:
    """Optionally defer exceptions and return a token instead

    When `__DEFER_STEP_ERRORS__` is set on the implementing class
    or instance, methods wrapped with this wrapper will
    catch and save their :python:`GSSError` exceptions and
    instead return the result token attached to the exception.

    The exception can be later retrived through :python:`_last_err`
    (and :python:`_last_tb` when Python 2 is in use).
    """

    try:
        return func(self, *args, **kwargs)
    except GSSError as e:
        defer_step_errors = getattr(self, '__DEFER_STEP_ERRORS__', False)
        if e.token is not None and defer_step_errors:
            self._last_err = e
            # skip the "return func" line above in the traceback
            tb = e.__traceback__.tb_next  # type: ignore[union-attr]
            self._last_err.__traceback__ = tb

            return e.token
        else:
            raise


@deco.decorator
def check_last_err(
    func: t.Callable,
    self: "SecurityContext",
    *args: t.Any,
    **kwargs: t.Any,
) -> t.Any:
    """Check and raise deferred errors before running the function

    This method checks :python:`_last_err` before running the wrapped
    function.  If present and not None, the exception will be raised
    with its original traceback.
    """

    if self._last_err is not None:
        try:
            raise self._last_err
        finally:
            self._last_err = None
    else:
        return func(self, *args, **kwargs)


class CheckLastError(type):
    """Check for a deferred error on all methods

    This metaclass applies the :python:`check_last_err` decorator
    to all methods not prefixed by '_'.

    Additionally, it enabled `__DEFER_STEP_ERRORS__` by default.
    """

    def __new__(
        cls,
        name: str,
        parents: t.Tuple[t.Type],
        attrs: t.Dict[str, t.Any],
    ) -> "CheckLastError":
        attrs['__DEFER_STEP_ERRORS__'] = True

        for attr_name in attrs:
            attr = attrs[attr_name]

            # wrap only methods
            if not isinstance(attr, types.FunctionType):
                continue

            if attr_name[0] != '_':
                attrs[attr_name] = check_last_err(attr)

        return super(CheckLastError, cls).__new__(cls, name, parents, attrs)



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import AcquireCredResult
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID


def acquire_cred_with_password(
    name: "Name",
    password: bytes,
    lifetime: t.Optional[int] = None,
    mechs: t.Optional[t.Iterable["OID"]] = None,
    usage: str = 'initiate',
) -> "AcquireCredResult":
    """Acquire credentials through provided password.

    This function is originally from Solaris and is not documented by either
    MIT or Heimdal.

    In general, it functions similarly to
    :func:`~gssapi.raw.creds.acquire_cred`.

    Args:
        name (~gssapi.raw.names.Name): the name to acquire credentials for
        password (bytes): the password used to acquire credentialss with
        lifetime (int): the lifetime for the credentials in seconds (or None
            for indefinite)
        mechs (~gssapi.raw.types.MechType): the desired mechanisms for which
            the credentials should work (or None for the default set)
        usage (str): usage type for credentials.  Possible values:
            'initiate' (default), 'accept', 'both' (failsafe).

    Returns:
        AcquireCredResult: the resulting credentials, the actual mechanisms
        with which they may be used, and their actual lifetime in seconds (or
        None for indefinite or not supported)

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.oids import OID


def set_neg_mechs(
    cred_handle: "Creds",
    mech_set: t.Iterable["OID"],
) -> None:
    """
    Specify the set of security mechanisms that may be negotiated with
    the credential identified by cred_handle.
    If more than one mechanism is specified in mech_set, the order in
    which those mechanisms are specified implies a relative preference.

    Args:
        cred_handle (Creds): credentials to set negotiable mechanisms for
        mech_set (~gssapi.raw.types.MechType): negotiable mechanisms to be set

    Returns:
        None

    Raises:
        ~gssapi.exceptions.GSSError
    """



"""
GGF Extensions

GGF provides extended credential and security context inquiry that allows
application to retrieve more information about the client's credentials and
security context. One common use case is to use
:meth:`inquire_sec_context_by_oid` to retrieve the "session" key that is
required by the SMB protocol for signing and encrypting a message.

Draft IETF document for these extensions can be found at
https://tools.ietf.org/html/draft-engert-ggf-gss-extensions-00
"""
import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.creds import Creds
    from gssapi.raw.oids import OID
    from gssapi.raw.sec_contexts import SecurityContext


def inquire_cred_by_oid(
    cred_handle: "Creds",
    desired_aspect: "OID",
) -> t.List[bytes]:
    """
    This method inspects a :class:`~gssapi.raw.creds.Creds` object for
    information specific to a particular desired aspect as an OID.

    Args:
        cred_handle (Creds): the Credentials to query
        desired_aspect (~gssapi.raw.oids.OID): the desired aspect of the
            Credentials to inquire about.

    Returns:
        list: A list of zero or more pieces of data (as bytes objects)

    Raises:
        ~gssapi.exceptions.GSSError
    """


def inquire_sec_context_by_oid(
    context: "SecurityContext",
    desired_aspect: "OID",
) -> t.List[bytes]:
    """
    This method inspects a :class:`~gssapi.raw.sec_contexts.SecurityContext`
    object for information specific to a particular desired aspect as an OID.

    This method can be used with the GSS_KRB5_INQ_SSPI_SESSION_KEY_OID OID to
    retrieve the required key that is used to derive the SMB/SAMBA signing and
    encryption keys.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the Security
            Context to query
        desired_aspect (~gssapi.raw.oids.OID): the desired aspect of the
            Security Context to inquire about.

    Returns:
        list: A list of zero or more pieces of data (as bytes objects)

    Raises:
        ~gssapi.exceptions.GSSError
    """


def set_sec_context_option(
    desired_aspect: "OID",
    context: "SecurityContext",
    value: t.Optional[bytes] = None,
) -> None:
    """
    This method is used to set a value for a specific OID of a
    :class:`~gssapi.raw.sec_contexts.SecurityContext` object. The OID and value
    to pass in depends on the mech the SecurityContext backs.

    An example of how this can be used would be to reset the NTLM crypto engine
    used in gss-ntlmssp. The OID that controls this value is
    '1.3.6.1.4.1.7165.655.1.3' and it takes it a byte value that represents
    an int32 where 1 resets the verifier handle and any other int resets the
    sender handle.

    Args:
        desired_aspect (~gssapi.raw.oids.OID): the desired aspect of the
            Security Context to set the value for.
        context (~gssapi.raw.sec_contexts.SecurityContext): the Security
            Context to set, or None to create a new context.
        value (bytes): the value to set on the desired aspect of the Security
            Context or None to send GSS_C_EMPTY_BUFFER.

    Returns:
        ~gssapi.raw.sec_contexts.SecurityContext: The output security context.

    Raises:
        ~gssapi.exceptions.GSSError
    """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID
    from gssapi.raw.named_tuples import (
        AcquireCredResult,
        AddCredResult,
        InquireCredResult,
        InquireCredByMechResult,
    )


class Creds:
    """
    GSSAPI Credentials
    """

    def __new__(
        cls,
        cpy: t.Optional["Creds"] = None,
    ) -> "Creds": ...


def acquire_cred(
    name: t.Optional["Name"],
    lifetime: t.Optional[int] = None,
    mechs: t.Optional[t.Iterable["OID"]] = None,
    usage: str = 'both',
) -> "AcquireCredResult":
    """Get GSSAPI credentials for the given name and mechanisms.

    This method gets GSSAPI credentials corresponding to the given name
    and mechanims.  The desired TTL and usage for the the credential may also
    be specified.

    Args:
        name (~gssapi.raw.names.Name): the name for which to acquire the
            credentials (or None for the "no name" functionality)
        lifetime (int): the lifetime in seconds for the credentials (or None
            for indefinite)
        mechs (~gssapi.raw.types.MechType): the desired mechanisms for which
            the credentials should work, or None for the default set
        usage (str): the usage type for the credentials: may be
            'initiate', 'accept', or 'both'

    Returns:
        AcquireCredResult: the resulting credentials, the actual mechanisms
        with which they may be used, and their actual lifetime in seconds (or
        None for indefinite or not supported)

    Raises:
        ~gssapi.exceptions.BadMechanismError
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
        ~gssapi.exceptions.ExpiredCredentialsError
        ~gssapi.exceptions.MissingCredentialsError
    """


def release_cred(
    creds: Creds,
) -> None:
    """
    release_cred(creds)
    Release GSSAPI Credentials.

    This method releases GSSAPI credentials.

    Warning:
        This method is deprecated.  Credentials are
        automatically freed by Python.

    Args:
        creds (Creds): the credentials in question

    Raises:
        ~gssapi.exceptions.MissingCredentialsError
    """


def add_cred(
    input_cred: Creds,
    name: "Name",
    mech: "OID",
    usage: str = 'initiate',
    init_lifetime: t.Optional[int] = None,
    accept_lifetime: t.Optional[int] = None,
    mutate_input: bool = False,
) -> "AddCredResult":
    """Add a credential element to a credential.

    This method can be used to either compose two credentials (i.e., original
    and new credential), or to add a new element to an existing credential.

    Args:
        input_cred (Creds): the set of credentials to which to add the new
            credentials
        name (~gssapi.raw.names.Name): name of principal to acquire a
            credential for
        mech (~gssapi.raw.types.MechType): the desired security mechanism
            (required).
        usage (str): usage type for credentials.  Possible values:
            'initiate' (default), 'accept', 'both' (failsafe).
        init_lifetime (int): lifetime of credentials for use in initiating
            security contexts in seconds (None for indefinite)
        accept_lifetime (int): lifetime of credentials for use in accepting
            security contexts in seconds (None for indefinite)
        mutate_input (bool): whether to mutate the input credentials (True)
            or produce a new set of credentials (False).  Defaults to False

    Returns:
        AddCredResult: the actual mechanisms with which the credentials may be
        used, the actual initiator TTL, and the actual acceptor TTL (None for
        either indefinite or not supported).  Note that the credentials may
        be set to None if mutate_input is set to True.

    Raises:
        ~gssapi.exceptions.BadMechanismError
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
        ~gssapi.exceptions.DuplicateCredentialsElementError
        ~gssapi.exceptions.ExpiredCredentialsError
        ~gssapi.exceptions.MissingCredentialsError
    """


def inquire_cred(
    creds: Creds,
    name: bool = True,
    lifetime: bool = True,
    usage: bool = True,
    mechs: bool = True,
) -> "InquireCredResult":
    """Inspect credentials for information.

    This method inspects a :class:`Creds` object for information.

    Args:
        creds (Creds): the credentials to inspect
        name (bool): get the Name associated with the credentials
        lifetime (bool): get the TTL for the credentials
        usage (bool): get the usage type of the credentials
        mechs (bool): the mechanims used with the credentials

    Returns:
        InquireCredResult: the information about the credentials,
        with unused fields set to None

    Raises:
        ~gssapi.exceptions.MissingCredentialsError
        ~gssapi.exceptions.InvalidCredentialsError
        ~gssapi.exceptions.ExpiredCredentialsError
    """


def inquire_cred_by_mech(
    creds: Creds,
    mech: "OID",
    name: bool = True,
    init_lifetime: bool = True,
    accept_lifetime: bool = True,
    usage: bool = True,
) -> "InquireCredByMechResult":
    """Inspect credentials for mechanism-specific information.

    This method inspects a :class:`Creds` object for information
    specific to a particular mechanism.  It functions similarly
    to :func:`inquire_cred`.

    Args:
        creds (Creds): the credentials to inspect
        mech (~gssapi.OID): the desired mechanism
        name (bool): get the Name associated with the credentials
        init_lifetime (bool): get the initiator TTL for the credentials (in
            seconds)
        accept_lifetime (bool): get the acceptor TTL for the credentials (in
            seconds)
        usage (bool): get the usage type of the credentials

    Returns:
        InquireCredByMechResult: the information about the credentials,
        with unused fields set to None

    Raises:
        ~gssapi.exceptions.MissingCredentialsError
        ~gssapi.exceptions.InvalidCredentialsError
    """



import typing as t

from gssapi.raw import chan_bindings as rchan_bindings
from gssapi.raw import sec_contexts as rsec_contexts
from gssapi.raw import message as rmessage
from gssapi.raw import named_tuples as tuples
from gssapi.raw import names as rnames
from gssapi.raw import oids as roids
from gssapi.raw.types import RequirementFlag, IntEnumFlagSet

import gssapi.exceptions as excs
from gssapi import _utils
from gssapi.names import Name
from gssapi.creds import Credentials


class SecurityContext(rsec_contexts.SecurityContext,
                      metaclass=_utils.CheckLastError):
    """A GSSAPI Security Context

    This class represents a GSSAPI security context that may be used
    with and/or returned by other GSSAPI methods.

    It inherits from the low-level GSSAPI
    :class:`~gssapi.raw.sec_contexts.SecurityContext` class,
    and thus may used with both low-level and high-level API methods.

    This class may be pickled and unpickled (the attached delegated
    credentials object will not be preserved, however).
    """

    def __new__(
        cls,
        base: t.Optional[rsec_contexts.SecurityContext] = None,
        token: t.Optional[bytes] = None,
        name: t.Optional[rnames.Name] = None,
        creds: t.Optional[Credentials] = None,
        lifetime: t.Optional[int] = None,
        flags: t.Optional[int] = None,
        mech: t.Optional[roids.OID] = None,
        channel_bindings: t.Optional[rchan_bindings.ChannelBindings] = None,
        usage: t.Optional[str] = None,
    ) -> "SecurityContext":

        if token is not None:
            base = rsec_contexts.import_sec_context(token)

        return t.cast("SecurityContext",
                      super(SecurityContext, cls).__new__(cls, base))

    def __init__(
        self,
        base: t.Optional[rsec_contexts.SecurityContext] = None,
        token: t.Optional[bytes] = None,
        name: t.Optional[rnames.Name] = None,
        creds: t.Optional[Credentials] = None,
        lifetime: t.Optional[int] = None,
        flags: t.Optional[int] = None,
        mech: t.Optional[roids.OID] = None,
        channel_bindings: t.Optional[rchan_bindings.ChannelBindings] = None,
        usage: t.Optional[str] = None,
    ) -> None:
        """
        The constructor creates a new security context, but does not begin
        the initiate or accept process.

        If the `base` argument is used, an existing
        :class:`~gssapi.raw.sec_contexts.SecurityContext` object from
        the low-level API is converted into a high-level object.

        If the `token` argument is passed, the security context is imported
        using the token.

        Otherwise, a new security context is created.

        If the `usage` argument is not passed, the constructor will attempt
        to detect what the appropriate usage is based on either the existing
        security context (if `base` or `token` are used) or the argument set.

        For a security context of the `initiate` usage, the `name` argument
        must be used, and the `creds`, `mech`, `flags`,
        `lifetime`, and `channel_bindings` arguments may be
        used as well.

        For a security context of the `accept` usage, the `creds` and
        `channel_bindings` arguments may optionally be used.
        """

        # NB(directxman12): _last_err must be set first
        self._last_err = None

        # determine the usage ('initiate' vs 'accept')
        if base is None and token is None:
            # this will be a new context
            if usage is not None:
                if usage not in ('initiate', 'accept'):
                    msg = "Usage must be either 'initiate' or 'accept'"
                    raise excs.UnknownUsageError(msg, obj="security context")

                self.usage = usage
            elif creds is not None and creds.usage != 'both':
                self.usage = creds.usage
            elif name is not None:
                # if we pass a name, assume the usage is 'initiate'
                self.usage = 'initiate'
            else:
                # if we don't pass a name, assume the usage is 'accept'
                self.usage = 'accept'

            # check for appropriate arguments
            if self.usage == 'initiate':
                # takes: creds?, target_name, mech?, flags?,
                #        channel_bindings?
                if name is None:
                    raise TypeError("You must pass the 'name' argument when "
                                    "creating an initiating security context")
                self._target_name = name
                self._mech = mech
                self._desired_flags = IntEnumFlagSet(RequirementFlag, flags)
                self._desired_lifetime = lifetime
            else:
                # takes creds?
                if (name is not None or flags is not None or
                        mech is not None or lifetime is not None):
                    raise TypeError("You must pass at most the 'creds' "
                                    "argument when creating an accepting "
                                    "security context")

            self._channel_bindings = channel_bindings
            self._creds = creds

            self._delegated_creds = None

        else:
            # we already have a context in progress, just inspect it
            # NB(directxman12): MIT krb5 refuses to inquire about a context
            # if it's partially established, so we have to check here

            try:
                if self.locally_initiated:
                    self.usage = 'initiate'
                else:
                    self.usage = 'accept'
            except excs.MissingContextError:
                msg = ("Cannot extract usage from a partially completed "
                       "context")
                raise excs.UnknownUsageError(msg, obj="security context")

        # This is to work around an MIT krb5 bug (see the `complete` property)
        self._complete: t.Optional[bool] = None

    # NB(directxman12): DO NOT ADD AN __del__ TO THIS CLASS -- it screws up
    #                   the garbage collector if _last_tb is still defined

    # TODO(directxman12): implement flag properties

    def get_signature(
        self,
        message: bytes,
    ) -> bytes:
        """Calculate the signature for a message.

        This method calculates the signature (called a MIC) for
        the given message, which may be then used with
        :meth:`verify_signature` to confirm the validity of the
        signature.  This is useful if you wish to transmit the
        message signature and message in your own format.

        Args:
            message (bytes): the input message

        Returns:
            bytes: the message signature

        Raises:
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
            ~gssapi.exceptions.BadQoPError
        """

        # TODO(directxman12): check flags?
        return rmessage.get_mic(self, message)

    def verify_signature(
        self,
        message: bytes,
        mic: bytes,
    ) -> int:
        """Verify the signature for a message.

        This method verifies that a signature (generated by
        :meth:`get_signature` is valid for the given message.

        If the signature is valid, the method will return.
        Otherwise, it will raise an error.

        Args:
            message (bytes): the message
            mic (bytes): the signature to verify

        Returns:
            int: the QoP used.

        Raises:
            ~gssapi.exceptions.BadMICError: the signature was not valid
            ~gssapi.exceptions.InvalidTokenError
            ~gssapi.exceptions.DuplicateTokenError
            ~gssapi.exceptions.ExpiredTokenError
            ~gssapi.exceptions.TokenTooLateError
            ~gssapi.exceptions.TokenTooEarlyError
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
        """

        return rmessage.verify_mic(self, message, mic)

    def wrap(
        self,
        message: bytes,
        encrypt: bool,
    ) -> tuples.WrapResult:
        """Wrap a message, optionally with encryption

        This wraps a message, signing it and optionally
        encrypting it.

        Args:
            message (bytes): the message to wrap
            encrypt (bool): whether or not to encrypt the message

        Returns:
            WrapResult: the wrapped message and details about it
            (e.g. whether encryption was used succesfully)

        Raises:
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
            ~gssapi.exceptions.BadQoPError
        """

        return rmessage.wrap(self, message, encrypt)

    def unwrap(
        self,
        message: bytes,
    ) -> tuples.UnwrapResult:
        """Unwrap a wrapped message.

        This method unwraps/unencrypts a wrapped message,
        verifying the signature along the way.

        Args:
            message (bytes): the message to unwrap/decrypt

        Returns:
            UnwrapResult: the unwrapped message and details about it
            (e.g. wheter encryption was used)

        Raises:
            ~gssapi.exceptions.InvalidTokenError
            ~gssapi.exceptions.BadMICError
            ~gssapi.exceptions.DuplicateTokenError
            ~gssapi.exceptions.ExpiredTokenError
            ~gssapi.exceptions.TokenTooLateError
            ~gssapi.exceptions.TokenTooEarlyError
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
        """

        return rmessage.unwrap(self, message)

    def encrypt(
        self,
        message: bytes,
    ) -> bytes:
        """Encrypt a message.

        This method wraps and encrypts a message, similarly to
        :meth:`wrap`.  The difference is that encryption is always
        used, and the method will raise an exception if this is
        not possible.  Additionally, this method simply returns
        the encrypted message directly.

        Args:
            message (bytes): the message to encrypt

        Returns:
            bytes: the encrypted message

        Raises:
            ~gssapi.exceptions.EncryptionNotUsed: the encryption could not be
                used
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
            ~gssapi.exceptions.BadQoPError
        """

        res = self.wrap(message, encrypt=True)

        if not res.encrypted:
            raise excs.EncryptionNotUsed("Wrapped message was not encrypted")

        return res.message

    def decrypt(
        self,
        message: bytes,
    ) -> bytes:
        """Decrypt a message.

        This method decrypts and unwraps a message, verifying the signature
        along the way, similarly to :meth:`unwrap`.  The difference is that
        this method will raise an exception if encryption was established
        by the context and not used, and simply returns the decrypted
        message directly.

        Args:
            message (bytes): the encrypted message

        Returns:
            bytes: the decrypted message

        Raises:
            ~gssapi.exceptions.EncryptionNotUsed: encryption was expected, but
                not used
            ~gssapi.exceptions.InvalidTokenError
            ~gssapi.exceptions.BadMICError
            ~gssapi.exceptions.DuplicateTokenError
            ~gssapi.exceptions.ExpiredTokenError
            ~gssapi.exceptions.TokenTooLateError
            ~gssapi.exceptions.TokenTooEarlyError
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
        """

        res = self.unwrap(message)

        if (not res.encrypted and
                self.actual_flags & RequirementFlag.confidentiality):
            raise excs.EncryptionNotUsed("The context was established with "
                                         "encryption, but unwrapped message "
                                         "was not encrypted",
                                         unwrapped_message=res.message)

        return res.message

    def get_wrap_size_limit(
        self,
        desired_output_size: int,
        encrypted: bool = True,
    ) -> int:
        """Calculate the maximum message size for a given wrapped message size.

        This method calculates the maximum input message size for a given
        maximum wrapped/encrypted message size.

        Args:
            desired_output_size (int): the maximum output message size
            encrypted (bool): whether or not encryption should be taken
                into account

        Returns:
            int: the maximum input message size

        Raises:
            ~gssapi.exceptions.MissingContextError
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.BadQoPError
        """

        return rmessage.wrap_size_limit(self, desired_output_size,
                                        encrypted)

    def process_token(
        self,
        token: bytes,
    ) -> None:
        """Process an output token asynchronously.

        This method processes an output token even when the security context
        was not expecting it.

        Warning:
            This method is deprecated.

        Args:
            token (bytes): the token to process

        Raises:
            ~gssapi.exceptions.InvalidTokenError
            ~gssapi.exceptions.MissingContextError
        """

        rsec_contexts.process_context_token(self, token)

    def export(self) -> bytes:
        """Export a security context.

        This method exports a security context, allowing it to be passed
        between processes.

        Returns:
            bytes: the exported security context

        Raises:
            ~gssapi.exceptions.ExpiredContextError
            ~gssapi.exceptions.MissingContextError
            ~gssapi.exceptions.OperationUnavailableError
        """

        return rsec_contexts.export_sec_context(self)

    _INQUIRE_ARGS = ('initiator_name', 'target_name', 'lifetime',
                     'mech', 'flags', 'locally_init', 'complete')

    @_utils.check_last_err
    def _inquire(
        self,
        **kwargs: bool,
    ) -> tuples.InquireContextResult:
        """Inspect the security context for information

        This method inspects the security context for information.

        If no keyword arguments are passed, all available information
        is returned.  Otherwise, only the keyword arguments that
        are passed and set to `True` are returned.

        Args:
            initiator_name (bool): get the initiator name for this context
            target_name (bool): get the target name for this context
            lifetime (bool): get the remaining lifetime, in seconds, for this
                context
            mech (bool): get the :class:`MechType` used by this context
            flags (bool): get the flags set on this context
            locally_init (bool): get whether this context was locally initiated
            complete (bool): get whether negotiation on this context has
                been completed

        Returns:
            InquireContextResult: the results of the inquiry, with unused
            fields set to None

        Raises:
            ~gssapi.exceptions.MissingContextError
        """
        if not kwargs:
            default_val = True
        else:
            default_val = False

        for arg in self._INQUIRE_ARGS:
            kwargs[arg] = kwargs.get(arg, default_val)

        res = rsec_contexts.inquire_context(self, **kwargs)

        if (kwargs.get('initiator_name', False) and
                res.initiator_name is not None):
            init_name = Name(res.initiator_name)
        else:
            init_name = None

        if (kwargs.get('target_name', False) and
                res.target_name is not None):
            target_name = Name(res.target_name)
        else:
            target_name = None

        return tuples.InquireContextResult(init_name, target_name,
                                           res.lifetime, res.mech,
                                           res.flags, res.locally_init,
                                           res.complete)

    @property
    def lifetime(self) -> int:
        """The amount of time for which this context remains valid"""
        return rsec_contexts.context_time(self)

    @property
    def delegated_creds(self) -> t.Optional[Credentials]:
        """The credentials delegated from the initiator to the acceptor

        .. warning::

            This value will not be preserved across picklings.  These should
            be separately exported and transferred.

        """
        return self._delegated_creds

    initiator_name = _utils.inquire_property(
        'initiator_name', 'The :class:`Name` of the initiator of this context')
    target_name = _utils.inquire_property(
        'target_name', 'The :class:`Name` of the target of this context')
    mech = _utils.inquire_property(
        'mech', 'The mechanism (:class:`MechType`) in use by this context')
    actual_flags = _utils.inquire_property(
        'flags', 'The flags set on this context')
    locally_initiated = _utils.inquire_property(
        'locally_init', 'Whether this context was locally intiated')

    @property
    @_utils.check_last_err
    def complete(self) -> bool:
        """Whether negotiation for this context has been completed"""
        # NB(directxman12): MIT krb5 has a bug where it refuses to
        #                   inquire about partially completed contexts,
        #                   so we can't just use `self._inquire` generally
        if self._started:
            complete = self._complete
            if complete is None:
                try:
                    complete = self._inquire(complete=True).complete
                except excs.MissingContextError:
                    return False
                else:
                    self._complete = complete

            return complete
        else:
            return False

    @_utils.catch_and_return_token
    def step(
        self,
        token: t.Optional[bytes] = None,
    ) -> t.Optional[bytes]:
        """Perform a negotation step.

        This method performs a negotiation step based on the usage type
        of this context.  If `__DEFER_STEP_ERRORS__` is set to True on
        the class, this method will return a token, even when exceptions
        would be thrown.  The generated exception will be thrown on the next
        method call or property lookup on the context.
        **This is the default behavior.**

        This method should be used in a while loop, as such:

        .. code-block:: python

           input_token = None
           try:
               while not ctx.complete:
                   output_token = ctx.step(input_token)
                   if not output_token:
                       break
                   input_token = send_and_receive(output_token)
           except GSSError as e:
                handle_the_issue()

        .. tip::

            Disabling `__DEFER_STEP_ERRORS__` is rarely necessary.
            When this method is used in a loop (as above),
            `__DEFER_STEP_ERRORS__` will ensure that you always
            send an error token when it's available,
            keeping the other end of the security context updated
            with the status of the negotiation.

        Args:
            token (bytes): the input token from the other participant's step

        Returns:
            bytes: the output token to send to the other participant

        Raises:
            ~gssapi.exceptions.InvalidTokenError
            ~gssapi.exceptions.InvalidCredentialsError
            ~gssapi.exceptions.MissingCredentialsError
            ~gssapi.exceptions.ExpiredCredentialsError
            ~gssapi.exceptions.BadChannelBindingsError
            ~gssapi.exceptions.BadMICError
            ~gssapi.exceptions.ExpiredTokenError: (initiate only)
            ~gssapi.exceptions.DuplicateTokenError
            ~gssapi.exceptions.MissingContextError
            ~gssapi.exceptions.BadNameTypeError: (initiate only)
            ~gssapi.exceptions.BadNameError: (initiate only)
            ~gssapi.exceptions.BadMechanismError
        """

        if self.usage == 'accept':
            return self._acceptor_step(token=token or b"")
        else:
            return self._initiator_step(token=token)

    def _acceptor_step(
        self,
        token: bytes,
    ) -> t.Optional[bytes]:
        res = rsec_contexts.accept_sec_context(token, self._creds,
                                               self, self._channel_bindings)

        if res.delegated_creds is not None:
            self._delegated_creds = Credentials(res.delegated_creds)
        else:
            self._delegated_creds = None

        self._complete = not res.more_steps

        return res.token

    def _initiator_step(
        self,
        token: t.Optional[bytes] = None,
    ) -> t.Optional[bytes]:
        res = rsec_contexts.init_sec_context(self._target_name, self._creds,
                                             self, self._mech,
                                             self._desired_flags,
                                             self._desired_lifetime,
                                             self._channel_bindings,
                                             token)

        self._complete = not res.more_steps

        return res.token

    # pickle protocol support
    def __reduce__(
        self,
    ) -> t.Tuple[t.Type["SecurityContext"], t.Tuple[None, bytes]]:
        # the unpickle arguments to new are (base=None, token=self.export())
        return (type(self), (None, self.export()))




import typing as t

from gssapi.raw import names as rname
from gssapi.raw import NameType
from gssapi.raw import named_tuples as tuples
from gssapi.raw import oids as roids
from gssapi import _utils

from collections.abc import MutableMapping, Iterable

rname_rfc6680 = _utils.import_gssapi_extension('rfc6680')
rname_rfc6680_comp_oid = _utils.import_gssapi_extension('rfc6680_comp_oid')


class Name(rname.Name):
    """A GSSAPI Name

    This class represents a GSSAPI name which may be used with and/or returned
    by other GSSAPI methods.

    It inherits from the low-level GSSAPI :class:`~gssapi.raw.names.Name`
    class, and thus may used with both low-level and high-level API methods.

    This class may be pickled and unpickled, as well as copied.

    The :class:`str` and :class:`bytes` methods may be used to retrieve the
    text of the name.

    Note:
        Name strings will be automatically converted to and from unicode
        strings as appropriate.  If a method is listed as returning a
        :class:`str` object, it will return a unicode string.

        The encoding used will be python-gssapi's current encoding, which
        defaults to UTF-8.
    """

    __slots__ = ('_attr_obj')

    def __new__(
        cls,
        base: t.Optional[t.Union[rname.Name, bytes, str]] = None,
        name_type: t.Optional[roids.OID] = None,
        token: t.Optional[bytes] = None,
        composite: bool = False,
    ) -> "Name":
        if token is not None:
            if composite:
                if rname_rfc6680 is None:
                    raise NotImplementedError(
                        "Your GSSAPI implementation does not support RFC 6680 "
                        "(the GSSAPI naming extensions)")

                if rname_rfc6680_comp_oid is not None:
                    base_name = rname.import_name(token,
                                                  NameType.composite_export)
                    displ_name = rname.display_name(base_name, name_type=True)
                    if displ_name.name_type == NameType.composite_export:
                        # NB(directxman12): there's a bug in MIT krb5 <= 1.13
                        # where GSS_C_NT_COMPOSITE_EXPORT doesn't trigger
                        # immediate import logic.  However, we can just use
                        # the normal GSS_C_NT_EXPORT_NAME in this case.
                        base_name = rname.import_name(token, NameType.export)
                else:
                    # NB(directxman12): some older versions of MIT krb5 don't
                    # have support for the GSS_C_NT_COMPOSITE_EXPORT, but do
                    # support composite tokens via GSS_C_NT_EXPORT_NAME.
                    base_name = rname.import_name(token, NameType.export)
            else:
                base_name = rname.import_name(token, NameType.export)
        elif isinstance(base, rname.Name):
            base_name = base
        else:
            if isinstance(base, str):
                base = base.encode(_utils._get_encoding())

            base_name = rname.import_name(
                base,  # type: ignore[arg-type]
                name_type)

        return t.cast("Name", super(Name, cls).__new__(cls, base_name))

    def __init__(
        self,
        base: t.Optional[t.Union[rname.Name, bytes, str]] = None,
        name_type: t.Optional[roids.OID] = None,
        token: t.Optional[bytes] = None,
        composite: bool = False,
    ) -> None:
        """
        The constructor can be used to "import" a name from a human readable
        representation, or from a token, and can also be used to convert a
        low-level :class:`gssapi.raw.names.Name` object into a high-level
        object.

        If a :class:`~gssapi.raw.names.Name` object from the low-level API
        is passed as the `base` argument, it will be converted into a
        high-level object.

        If the `token` argument is used, the name will be imported using
        the token.  If the token was exported as a composite token,
        pass `composite=True`.

        Otherwise, a new name will be created, using the `base` argument as
        the human-readable string and the `name_type` argument to denote the
        name type.

        Raises:
            ~gssapi.exceptions.BadNameTypeError
            ~gssapi.exceptions.BadNameError
            ~gssapi.exceptions.BadMechanismError
        """

        self._attr_obj: t.Optional[_NameAttributeMapping]

        if rname_rfc6680 is not None:
            self._attr_obj = _NameAttributeMapping(self)
        else:
            self._attr_obj = None

    def __str__(self) -> str:
        return bytes(self).decode(_utils._get_encoding())

    def __unicode__(self) -> str:
        # Python 2 -- someone asked for unicode
        return self.__bytes__().decode(_utils._get_encoding())

    def __bytes__(self) -> bytes:
        # Python 3 -- someone asked for bytes
        return rname.display_name(self, name_type=False).name

    def display_as(
        self,
        name_type: roids.OID,
    ) -> str:
        """
        Display this name as the given name type.

        This method attempts to display the current :class:`Name`
        using the syntax of the given :class:`~gssapi.raw.types.NameType`, if
        possible.

        Warning:

            In MIT krb5 versions below 1.13.3, this method can segfault if
            the name was not *originally* created with a `name_type` that was
            not ``None`` (even in cases when a ``name_type``
            is later "added", such as via :meth:`canonicalize`).
            **Do not use this method unless you are sure the above
            conditions can never happen in your code.**

        Warning:

            In addition to the above warning, current versions of MIT krb5 do
            not actually fully implement this method, and it may return
            incorrect results in the case of canonicalized names.

        :requires-ext:`rfc6680`

        Args:
            name_type (~gssapi.OID): the :class:`~gssapi.raw.types.NameType` to
                use to display the given name

        Returns:
            str: the displayed name

        Raises:
            ~gssapi.exceptions.OperationUnavailableError
        """

        if rname_rfc6680 is None:
            raise NotImplementedError("Your GSSAPI implementation does not "
                                      "support RFC 6680 (the GSSAPI naming "
                                      "extensions)")
        return rname_rfc6680.display_name_ext(self, name_type).decode(
            _utils._get_encoding())

    @property
    def name_type(self) -> t.Optional[roids.OID]:
        """The :class:`~gssapi.raw.types.NameType` of this name"""
        return rname.display_name(self, name_type=True).name_type

    def __eq__(
        self,
        other: object,
    ) -> bool:
        if not isinstance(other, rname.Name):
            # maybe something else can compare this
            # to other classes, but we certainly can't
            return NotImplemented
        else:
            return rname.compare_name(self, other)

    def __ne__(
        self,
        other: object,
    ) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        disp_res = rname.display_name(self, name_type=True)
        return "Name({name!r}, {name_type})".format(
            name=disp_res.name, name_type=disp_res.name_type)

    def export(
        self,
        composite: bool = False,
    ) -> bytes:
        """Export this name as a token.

        This method exports the name into a byte string which can then be
        imported by using the `token` argument of the constructor.

        Args:
            composite (bool): whether or not use to a composite token --
                :requires-ext:`rfc6680`

        Returns:
            bytes: the exported name in token form

        Raises:
            ~gssapi.exceptions.MechanismNameRequiredError
            ~gssapi.exceptions.BadNameTypeError
            ~gssapi.exceptions.BadNameError
        """

        if composite:
            if rname_rfc6680 is None:
                raise NotImplementedError("Your GSSAPI implementation does "
                                          "not support RFC 6680 (the GSSAPI "
                                          "naming extensions)")

            return rname_rfc6680.export_name_composite(self)
        else:
            return rname.export_name(self)

    def canonicalize(
        self,
        mech: roids.OID
    ) -> "Name":
        """Canonicalize a name with respect to a mechanism.

        This method returns a new :class:`Name` that is canonicalized according
        to the given mechanism.

        Args:
            mech (~gssapi.OID): the :class:`MechType` to use

        Returns:
            Name: the canonicalized name

        Raises:
            ~gssapi.exceptions.BadMechanismError
            ~gssapi.exceptions.BadNameTypeError
            ~gssapi.exceptions.BadNameError
        """

        return type(self)(rname.canonicalize_name(self, mech))

    def __copy__(self) -> "Name":
        return type(self)(rname.duplicate_name(self))

    def __deepcopy__(
        self,
        memo: t.Dict,
    ) -> "Name":
        return type(self)(rname.duplicate_name(self))

    def _inquire(
        self,
        **kwargs: t.Any,
    ) -> tuples.InquireNameResult:
        """Inspect this name for information.

        This method inspects the name for information.

        If no keyword arguments are passed, all available information
        is returned.  Otherwise, only the keyword arguments that
        are passed and set to `True` are returned.

        Args:
            mech_name (bool): get whether this is a mechanism name,
                and, if so, the associated mechanism
            attrs (bool): get the attributes names for this name

        Returns:
            InquireNameResult: the results of the inquiry, with unused
                fields set to None

        Raises:
            ~gssapi.exceptions.GSSError
        """

        if rname_rfc6680 is None:
            raise NotImplementedError("Your GSSAPI implementation does not "
                                      "support RFC 6680 (the GSSAPI naming "
                                      "extensions)")

        if not kwargs:
            default_val = True
        else:
            default_val = False

        attrs = kwargs.get('attrs', default_val)
        mech_name = kwargs.get('mech_name', default_val)

        return rname_rfc6680.inquire_name(self, mech_name=mech_name,
                                          attrs=attrs)

    @property
    def is_mech_name(self) -> bool:
        """Whether or not this name is a mechanism name
        (:requires-ext:`rfc6680`)
        """
        return self._inquire(mech_name=True).is_mech_name

    @property
    def mech(self) -> roids.OID:
        """The mechanism associated with this name (:requires-ext:`rfc6680`)
        """
        return self._inquire(mech_name=True).mech

    @property
    def attributes(self) -> t.Optional[MutableMapping]:
        """The attributes of this name (:requires-ext:`rfc6680`)

        The attributes are presenting in the form of a
        :class:`~collections.abc.MutableMapping` (a dict-like object).

        Retrieved values will always be in the form of :class:`frozenset`.

        When assigning values, if iterables are used, they be considered to be
        the set of values for the given attribute.  If a non-iterable is used,
        it will be considered a single value, and automatically wrapped in an
        iterable.

        Note:
            String types (includes :class:`bytes`) are not considered to
            be iterables in this case.
        """
        if self._attr_obj is None:
            raise NotImplementedError("Your GSSAPI implementation does not "
                                      "support RFC 6680 (the GSSAPI naming "
                                      "extensions)")

        return self._attr_obj


class _NameAttributeMapping(MutableMapping):

    """Provides dict-like access to RFC 6680 Name attributes."""
    def __init__(
        self,
        name: Name,
    ) -> None:
        self._name = name

    def __getitem__(
        self,
        key: t.Union[bytes, str],
    ) -> tuples.GetNameAttributeResult:
        if isinstance(key, str):
            key = key.encode(_utils._get_encoding())

        res = rname_rfc6680.get_name_attribute(  # type: ignore[union-attr]
            self._name, key)
        res = t.cast(tuples.GetNameAttributeResult, res)

        return tuples.GetNameAttributeResult(list(res.values),
                                             list(res.display_values),
                                             res.authenticated,
                                             res.complete)

    def __setitem__(
        self,
        key: t.Union[bytes, str],
        value: t.Union[
            tuples.GetNameAttributeResult, t.Tuple[bytes, bool], bytes
        ],
    ) -> None:
        if isinstance(key, str):
            key = key.encode(_utils._get_encoding())

        rname_rfc6680.delete_name_attribute(  # type: ignore[union-attr]
            self._name, key)

        attr_value: t.List[bytes]
        if isinstance(value, tuples.GetNameAttributeResult):
            complete = value.complete
            attr_value = value.values
        elif isinstance(value, tuple) and len(value) == 2:
            complete = value[1]
            attr_value = [value[0]]
        else:
            complete = False

        if (isinstance(value, (str, bytes)) or
                not isinstance(value, Iterable)):
            # NB(directxman12): this allows us to easily assign a single
            # value, since that's a common case
            attr_value = [value]

        rname_rfc6680.set_name_attribute(  # type: ignore[union-attr]
            self._name, key, attr_value, complete=complete)

    def __delitem__(
        self,
        key: t.Union[bytes, str],
    ) -> None:
        if isinstance(key, str):
            key = key.encode(_utils._get_encoding())

        rname_rfc6680.delete_name_attribute(  # type: ignore[union-attr]
            self._name, key)

    def __iter__(self) -> t.Iterator[bytes]:
        return iter(self._name._inquire(attrs=True).attrs)

    def __len__(self) -> int:
        return len(self._name._inquire(attrs=True).attrs)



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import WrapResult, UnwrapResult
    from gssapi.sec_contexts import SecurityContext


def get_mic(
    context: "SecurityContext",
    message: bytes,
    qop: t.Optional[int] = None,
) -> bytes:
    """Generate a MIC for a message.

    This method generates a Message Integrity Check token for the
    given message.  This can be separately trasmitted to the other
    entity, unlike wrap, which bundles the MIC and the message
    together.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (bytes): the message for which to generate the MIC
        qop (int): the requested Quality of Protection
            (or None to use the default)

    Returns:
        bytes: the generated MIC token

    Raises:
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.BadQoPError
    """


def verify_mic(
    context: "SecurityContext",
    message: bytes,
    token: bytes,
) -> int:
    """Verify that a MIC matches a message.

    This method verifies that the given MIC matches the given message.
    If the MIC does not match the given message, an exception will
    be raised.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (bytes): the message in question
        token (bytes): the MIC token in question

    Returns:
        int: the QoP used.

    Raises:
        ~gssapi.exceptions.InvalidTokenError
        ~gssapi.exceptions.BadMICError
        ~gssapi.exceptions.DuplicateTokenError
        ~gssapi.exceptions.ExpiredTokenError
        ~gssapi.exceptions.TokenTooLateError
        ~gssapi.exceptions.TokenTooEarlyError
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.MissingContextError
    """


def wrap_size_limit(
    context: "SecurityContext",
    output_size: int,
    confidential: bool = True,
    qop: t.Optional[int] = None,
) -> int:
    """Calculate the max message size.

    This method calculates the unwrapped/unencrypted message size for
    the given maximum wrapped/encrypted message size.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        output_size (int): the maximum desired wrapped/encrypted message size
        confidential (bool): whether or not confidentiality is being used
        qop (int): the QoP that will be when you actually call wrap
            (or None for the default QoP)

    Returns:
        int: the maximum unencrypted/unwrapped message size

    Raises:
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.BadQoPError
    """


def wrap(
    context: "SecurityContext",
    message: bytes,
    confidential: bool = True,
    qop: t.Optional[int] = None,
) -> "WrapResult":
    """Wrap/Encrypt a message.

    This method wraps or encrypts a message (depending on the value
    of confidential) with the given Quality of Protection.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (bytes): the message to wrap or encrypt
        confidential (bool): whether or not to encrypt the message (True),
            or just wrap it with a MIC (False)
        qop (int): the desired Quality of Protection
            (or None for the default QoP)

    Returns:
        WrapResult: the wrapped/encrypted message, and whether or not
            encryption was actually used

    Raises:
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.MissingContextError
        ~gssapi.exceptions.BadQoPError
    """


def unwrap(
    context: "SecurityContext",
    message: bytes,
) -> "UnwrapResult":
    """Unwrap/Decrypt a message.

    This method unwraps or decrypts a message, depending
    on whether the sender used confidentiality.

    Args:
        context (~gssapi.raw.sec_contexts.SecurityContext): the current
            security context
        message (bytes): the message to unwrap/decrypt

    Returns:
        UnwrapResult: the unwrapped/decrypted message, whether or on
            encryption was used, and the QoP used

    Raises:
        ~gssapi.exceptions.InvalidTokenError
        ~gssapi.exceptions.BadMICError
        ~gssapi.exceptions.DuplicateTokenError
        ~gssapi.exceptions.ExpiredTokenError
        ~gssapi.exceptions.TokenTooLateError
        ~gssapi.exceptions.TokenTooEarlyError
        ~gssapi.exceptions.ExpiredContextError
        ~gssapi.exceptions.MissingContextError
    """



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import DisplayNameResult
    from gssapi.raw.oids import OID

class Name:
    """
    A GSSAPI Name
    """

    def __new__(
        cls,
        cpy: t.Optional["Name"] = None,
    ) -> "Name": ...


def import_name(
    name: bytes,
    name_type: t.Optional["OID"] = None,
) -> Name:
    """Convert a string and a name type into a GSSAPI name.

    This method takes a string name and a name type and converts
    them into a GSSAPI :class:`Name`.

    Args:
        name (~gssapi.raw.names.Name): the string version of the name
        name_type (~gssapi.raw.types.MechType): the type of this name

    Returns:
        Name: the GSSAPI version of the name

    Raises:
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
        ~gssapi.exceptions.BadMechanismError
    """


def display_name(
    name: Name,
    name_type: bool = True,
) -> "DisplayNameResult":
    """Convert a GSSAPI name into its components.

    This method converts a GSSAPI :class:`Name` back into its
    text form.  If ``name_type`` is True, it also attempts to
    retrieve the :class:`~gssapi.raw.types.NameType` of the name (otherwise the
    returned name type will be ``None``).

    Args:
        name (~gssapi.raw.names.Name): the name in question
        name_type (~gssapi.raw.types.MechType): whether or not to retrieve the
            name type

    Returns:
        DisplayNameResult: the text part of the name and its type

    Raises:
        ~gssapi.exceptions.BadNameError
    """


def compare_name(
    name1: Name,
    name2: Name,
) -> bool:
    """Check two GSSAPI names to see if they are the same.

    This method compares two GSSAPI names, checking to
    see if they are equivalent.

    Args:
        name1 (~gssapi.raw.names.Name): the first name to compare
        name2 (~gssapi.raw.names.Name): the second name to compare

    Returns:
        bool: whether or not the names are equal

    Raises:
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
    """


def export_name(
    name: Name,
) -> bytes:
    """Export a GSSAPI name.

    This method "produces a canonical contigous string representation
    of a mechanism name, suitable for direct comparison for use in
    authorization functions".

    The input name must be a valid GSSAPI mechanism name, as generated by
    :func:`canonicalize_name` or
    :func:`~gssapi.raw.sec_contexts.accept_sec_context`.

    Args:
        name (~gssapi.raw.names.Name): the name to export

    Returns:
        bytes: the exported name

    Raises:
        ~gssapi.exceptions.MechanismNameRequiredError
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
    """


def canonicalize_name(
    name: Name,
    mech: "OID",
) -> Name:
    """Canonicalize an arbitrary GSSAPI Name into a Mechanism Name

    This method turns any GSSAPI name into a "mechanism name" --
    a full form name specific to a mechanism.

    Args:
        name (~gssapi.raw.names.Name): the name to canonicalize
        mech (~gssapi.raw.types.MechType): the mechanism type to use to
            canonicalize the name

    Returns:
        Name: a canonicalized version of the input name

    Raises:
        ~gssapi.exceptions.BadMechanismError
        ~gssapi.exceptions.BadNameTypeError
        ~gssapi.exceptions.BadNameError
    """


def duplicate_name(
    name: Name,
) -> Name:
    """Duplicate a GSSAPI name.

    Args:
        name (~gssapi.raw.names.Name): the name to duplicate

    Returns:
        Name: a duplicate of the input name

    Raises:
        ~gssapi.exceptions.BadNameError
    """


def release_name(
    name: Name,
) -> None:
    """Release a GSSAPI name.

    This method frees a GSSAPI :class:`Name`.
    You probably won't have to do this.

    Warning:
        This method is deprecated.  Names are
        automatically freed by Python.

    Args:
        name (~gssapi.raw.names.Name): the name in question

    Raises:
        ~gssapi.exceptions.BadNameError
    """



import typing as t

from enum import EnumMeta


_extra_values: t.Dict[str, t.Dict[str, t.Any]] = {}


def register_value(
    cl_str: str,
    name: str,
    value: t.Any,
) -> None:
    _extra_values[cl_str] = _extra_values.get(cl_str, {})
    _extra_values[cl_str][name] = value


class ExtendableEnum(EnumMeta):
    def __new__(
        metacl,
        name: str,
        bases: t.Tuple[t.Type],
        classdict: t.Dict[str, t.Any],
    ) -> "ExtendableEnum":
        extra_vals = _extra_values.get(name)

        if extra_vals is not None:
            for extra_name, extra_val in list(extra_vals.items()):
                if extra_name in classdict:
                    raise AttributeError(
                        "Enumeration extensions cannot override existing "
                        "enumeration members")
                else:
                    classdict[extra_name] = extra_val

        return super(ExtendableEnum, metacl).__new__(
            metacl,
            name,
            bases,
            classdict,  # type: ignore[arg-type] # Uses private explicit type
        )



"""
Using GSSAPI on Windows requires having an installation of Kerberos for Windows
(KfW) available in the user's PATH. This module should be imported before
anything else to check for that installation, add it to the PATH if necessary,
and throw any errors before they manifest as cryptic missing DLL errors later
down the import tree.
"""

import os
import shutil
import sys
import ctypes

#: Path to normal KfW installed bin folder
KFW_BIN = os.path.join(
    os.environ.get('ProgramFiles', r'C:\Program Files'),
    'MIT', 'Kerberos', 'bin',
)
#: Download location for KfW
KFW_DL = "https://web.mit.edu/KERBEROS/dist"

# Mypy needs to run on both Win and non-Win so the missing attribute will fire
# on non-Win and Win will fire with unused ignore. Instead just cache the attr
# by name and use it as needed.
ADD_DLL_DIR = getattr(os, "add_dll_directory", None)
CTYPES_WIN_DLL = getattr(ctypes, "WinDLL", ctypes.CDLL)


def _add_dll_directory(path: str) -> None:
    if ADD_DLL_DIR:
        ADD_DLL_DIR(path)


def kfw_available() -> bool:
    """Return if the main GSSAPI DLL for KfW can be loaded"""
    try:  # to load the main GSSAPI DLL
        if sys.maxsize > 2**32:
            CTYPES_WIN_DLL('gssapi64.dll')
        else:
            CTYPES_WIN_DLL('gssapi32.dll')
    except OSError:  # DLL is not in PATH
        return False
    else:  # DLL is in PATH, everything should work
        return True


def error_not_found() -> None:
    """Raise an OSError detailing that KfW is missing and how to get it"""
    raise OSError(
        "Could not find KfW installation. Please download and install "
        "the 64bit Kerberos for Windows MSI from %s and ensure the "
        "'bin' folder (%s) is in your PATH."
        % (KFW_DL, KFW_BIN)
    )


def configure_windows() -> None:
    """
    Validate that KfW appears to be installed correctly and add it to the
    DLL directories/PATH if necessary. In the case that it can't be located,
    raise an error.
    """
    if kfw_available():
        return  # All set, necessary DLLs should be available

    if os.path.exists(KFW_BIN):  # In standard location
        try:  # to use Python 3.8's DLL handling
            _add_dll_directory(KFW_BIN)
        except AttributeError:  # <3.8, use PATH
            os.environ['PATH'] += os.pathsep + KFW_BIN
        if kfw_available():
            return

    # Check if kinit is in the PATH which should lead us to the bin folder
    kinit_path = shutil.which('kinit')  # KfW provided binary
    if kinit_path:  # Non-standard install location
        try:  # Most likely >=3.8, otherwise it would have been found already
            _add_dll_directory(os.path.dirname(kinit_path))
        except AttributeError:  # <3.8, corrupted installation?
            pass
        else:
            if kfw_available():
                return

    error_not_found()


if os.name == 'nt':  # Make sure we have the required DLLs
    configure_windows()



import typing as t

class OID:
    """
    A GSSAPI OID

    A new OID may be created by passing the `elements` argument
    to the constructor.  The `elements` argument should be a
    :class:`bytes` consisting of the BER-encoded values in the OID.

    To retrieve the underlying bytes, use the :class:`bytes`
    function in Python 3.

    This object is hashable, and may be compared using equality
    operators.
    """

    def __new__(
        cls,
        cpy: t.Optional["OID"] = None,
        elements: t.Optional[bytes] = None,
    ) -> "OID": ...

    @classmethod
    def from_int_seq(
        cls,
        integer_sequence: t.Union[str, t.Iterable[int]],
    ) -> "OID":
        """Create a OID from a sequence of integers.

        This method creates an OID from a sequence of integers.
        The sequence can either be in dotted form as a string,
        or in list form.

        This method is not for BER-encoded byte strings, which
        can be passed directly to the OID constructor.

        Args:
            integer_sequence: either a list of integers or
                a string in dotted form

        Returns:
            OID: the OID represented by the given integer sequence

        Raises:
            ValueError: the sequence is less than two elements long
        """

    @property
    def dotted_form(self) -> str: ...



import typing as t

if t.TYPE_CHECKING:
    from gssapi.raw.named_tuples import (
        GetNameAttributeResult,
        InquireNameResult,
    )
    from gssapi.raw.names import Name
    from gssapi.raw.oids import OID


def display_name_ext(
    name: "Name",
    name_type: "OID",
) -> bytes:
    """Display the given Name using the given name type.

    This method attempts to display the given Name using the syntax of
    the given name type.  If this is not possible, an appropriate error
    will be raised.

    Args:
        name (~gssapi.raw.names.Name): the name to display
        name_type (~gssapi.OID): the name type (see NameType) to use to
            display the given name

    Returns:
        bytes: the displayed name

    Raises:
        ~gssapi.exceptions.OperationUnavailableError: the given name could not
            be displayed using the given name type
    """


def inquire_name(
    name: "Name",
    mech_name: bool = True,
    attrs: bool = True,
) -> "InquireNameResult":
    """Get information about a Name.

    This method retrieves information about the given name, including
    the set of attribute names for the given name, as well as whether or
    not the name is a mechanism name.  Additionally, if the given name is
    a mechanism name, the associated mechansim is returned as well.

    Args:
        name (~gssapi.raw.names.Name): the name about which to inquire
        mech_name (bool): whether or not to retrieve if this name
            is a mech_name (and the associate mechanism)
        attrs (bool): whether or not to retrieve the attribute name list

    Returns:
        InquireNameResult: the set of attribute names for the given name,
        whether or not the name is a Mechanism Name, and potentially
        the associated mechanism if it is a Mechanism Name

    Raises:
        ~gssapi.exceptions.GSSError
    """


def set_name_attribute(
    name: "Name",
    attr: bytes,
    value: t.Iterable[bytes],
    complete: bool = False,
) -> None:
    """Set the value(s) of a name attribute.

    This method sets the value(s) of the given attribute on the given name.

    Note that this functionality more closely matches the pseudo-API
    presented in RFC 6680, not the C API (which uses multiple calls to
    add multiple values).  However, multiple calls to this method will
    continue adding values, so :func:`delete_name_attribute` must be
    used in between calls to "clear" the values.

    Args:
        name (~gssapi.raw.names.Name): the Name on which to set the attribute
        attr (bytes): the name of the attribute
        value (list): a list of bytes objects to use as the value(s)
        complete (bool): whether or not to mark this attribute's value
            set as being "complete"

    Raises:
        ~gssapi.exceptions.OperationUnavailableError: the given attribute name
            is unknown or could not be set
    """


def get_name_attribute(
    name: "Name",
    attr: bytes,
    more: t.Optional[int] = None,
) -> "GetNameAttributeResult":
    """Get the value(s) of a name attribute.

    This method retrieves the value(s) of the given attribute
    for the given Name.

    Note that this functionality matches pseudo-API presented
    in RFC 6680, not the C API (which uses a state variable and
    multiple calls to retrieve multiple values).

    Args:
        name (~gssapi.raw.names.Name): the Name from which to get the attribute
        attr (bytes): the name of the attribute

    Returns:
        GetNameAttributeResult: the raw version of the value(s),
        the human-readable version of the value(s), whether
        or not the attribute was authenticated, and whether or
        not the attribute's value set was marked as complete

    Raises:
        ~gssapi.exceptions.OperationUnavailableError: the given attribute is
            unknown or unset
    """


def delete_name_attribute(
    name: "Name",
    attr: bytes,
) -> None:
    """Remove an attribute from a name.

    This method removes an attribute from a Name.  This method may be
    used before :func:`set_name_attribute` clear the values of an attribute
    before setting a new value (making the latter method work like a 'set'
    operation instead of an 'add' operation).

    Note that the removal of certain attributes may not be allowed.

    Args:
        name (~gssapi.raw.names.Name): the name to remove the attribute from
        attr (bytes): the name of the attribute

    Raises:
        ~gssapi.exceptions.OperationUnavailableError
        ~gssapi.exceptions.UnauthorizedError
    """


def export_name_composite(
    name: "Name",
) -> bytes:
    """Export a name, preserving attribute information.

    This method functions similarly to :func:`~gssapi.raw.names.export_name`,
    except that it preserves attribute information.  The resulting bytes may be
    imported using :func:`~gssapi.raw.names.import_name` with the
    :attr:`~gssapi.raw.types.NameType.composite_export` name type.

    Note:
        Some versions of MIT Kerberos require you to either canonicalize a name
        once it has been imported with composite-export name type, or to import
        using the normal export name type.

    Args:
        name (~gssapi.raw.names.Name): the name to export

    Returns:
        bytes: the exported composite name

    Raises:
        ~gssapi.exceptions.GSSError
    """



"""High-Level GSSAPI Bindings

The high-level API contains three main classes, which represent
the primary abstractions that GSSAPI provides:

    Name (see gssapi.names)

    Credentials (see gssapi.creds)

    SecurityContext (see gssapi.sec_contexts)

Additionally, a number of helper classes shared with the low-level API
exist as well:

    Enums (see gssapi.raw.types) --
        NameType, RequirementFlag, AddressType, MechType

    IntEnumFlagSet (see gssapi.raw.types)

    OID (see gssapi.raw.oids)

Note:

    Classes in the high-level API inherit from the corresponding
    classes in the low-level API, and thus may be passed in to
    low-level API functions.
"""

import gssapi._win_config  # noqa

from gssapi.raw.types import NameType, RequirementFlag, AddressType  # noqa
from gssapi.raw.types import MechType, IntEnumFlagSet  # noqa
from gssapi.raw.oids import OID  # noqa

from gssapi.creds import Credentials  # noqa
from gssapi.names import Name  # noqa
from gssapi.sec_contexts import SecurityContext  # noqa
from gssapi.mechs import Mechanism  # noqa

from gssapi._utils import set_encoding  # noqa

__all__ = [
    'AddressType',
    'Credentials',
    'IntEnumFlagSet',
    'Mechanism',
    'MechType',
    'Name',
    'NameType',
    'OID',
    'RequirementFlag',
    'SecurityContext',
    'set_encoding',
]
